"""Tests for Ajax view permissions in the timeline app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions import ProjectPermissionTestBase

from timeline.tests.test_models import (
    TimelineEventMixin,
    TimelineEventStatusMixin,
)


# Local constants
APP_NAME_PR = 'projectroles'


class TestProjectEventDetailAjaxView(
    TimelineEventMixin, TimelineEventStatusMixin, ProjectPermissionTestBase
):
    """Tests for ProjectEventDetailAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            self.project, APP_NAME_PR, self.user_owner, 'project_create'
        )
        self.make_event_status(self.event, 'OK')
        self.url = reverse(
            'timeline:ajax_detail_project',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test ProjectEventDetailAjaxView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.user_no_roles, 200)
        self.assert_response(self.url, self.anonymous, 403)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)

    def test_get_classified(self):
        """Test GET with classified event"""
        self.event.classified = True
        self.event.save()
        good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_classified_anon(self):
        """Test GET with classified event and anonymous access"""
        self.event.classified = True
        self.event.save()
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 403)


class TestSiteEventDetailAjaxView(
    TimelineEventMixin, TimelineEventStatusMixin, ProjectPermissionTestBase
):
    """Tests for SiteEventDetailAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            None, APP_NAME_PR, self.user_owner, 'test_event'
        )
        self.make_event_status(self.event, 'OK')
        self.regular_user = self.make_user('regular_user')
        self.url = reverse(
            'timeline:ajax_detail_site',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )

    def test_get(self):
        """Test SiteEventDetailAjaxView GET"""
        good_users = [self.superuser, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, self.anonymous, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        good_users = [self.superuser, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(self.url, self.anonymous, 403)

    def test_get_classified(self):
        """Test GET with classified event"""
        self.event.classified = True
        self.event.save()
        bad_users = [self.regular_user, self.anonymous]
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, bad_users, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_classified_anon(self):
        """Test GET with classified event and anonymous access"""
        self.event.classified = True
        self.event.save()
        self.assert_response(self.url, self.no_role_users, 403)


class TestProjectEventExtraAjaxView(
    TimelineEventMixin, TimelineEventStatusMixin, ProjectPermissionTestBase
):
    """Tests for ProjectEventExtraAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            self.project, APP_NAME_PR, self.user_owner, 'project_create'
        )
        self.make_event_status(self.event, 'OK')
        self.url = reverse(
            'timeline:ajax_extra_project',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        self.bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test ProjectEventExtraAjaxView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test SiteEventDetailAjaxView permissions with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 403)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)

    def test_get_block(self):
        """Test GET with project access block"""
        self.set_access_block(self.project)
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, self.non_superusers, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)

    def test_get_classified(self):
        """Test GET with classified event"""
        self.event.classified = True
        self.event.save()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_classified_anon(self):
        """Test GET with classified event and anonymous access"""
        self.event.classified = True
        self.event.save()
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)


class TestSiteEventExtraAjaxView(
    TimelineEventMixin, TimelineEventStatusMixin, ProjectPermissionTestBase
):
    """Tests for SiteEventExtraAjaxView permissions"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            None,
            APP_NAME_PR,
            self.user_owner,
            'test_event',
            extra_data={'example_data': 'example_extra_data'},
        )
        self.event_status = self.make_event_status(self.event, 'OK')
        self.regular_user = self.make_user('regular_user')
        self.url = reverse(
            'timeline:ajax_extra_site',
            kwargs={'timelineevent': self.event.sodar_uuid},
        )

    def test_get(self):
        """Test SiteEventExtraAjaxView GET"""
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, [self.regular_user, self.anonymous], 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(self.url, self.anonymous, 403)

    def test_get_classified(self):
        """Test GET with classified event"""
        self.event.classified = True
        self.event.save()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, [self.regular_user, self.anonymous], 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_classified_anon(self):
        """Test GET with classified event and anonymous access"""
        self.event.classified = True
        self.event.save()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, [self.regular_user, self.anonymous], 403)


class TestEventStatusExtraAjaxViewProject(
    TimelineEventMixin, TimelineEventStatusMixin, ProjectPermissionTestBase
):
    """Tests for EventStatusExtraAjaxView permissions with a project event"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            self.project, APP_NAME_PR, self.user_owner, 'project_create'
        )
        self.event_status = self.make_event_status(
            self.event, 'OK', extra_data={'example_data': 'example_extra_data'}
        )
        self.url = reverse(
            'timeline:ajax_extra_status',
            kwargs={'eventstatus': self.event_status.sodar_uuid},
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_owner,
            self.user_delegate,
        ]
        self.bad_users = [
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_contributor,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
            self.anonymous,
        ]

    def test_get(self):
        """Test EventStatusExtraAjaxView GET"""
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)

    def test_get_classified(self):
        """Test GET with classified event"""
        self.event.classified = True
        self.event.save()
        self.assert_response(self.url, self.good_users, 200)
        self.assert_response(self.url, self.bad_users, 403)
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.no_role_users, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_classified_anon(self):
        """Test GET with classified event and anonymous access"""
        self.event.classified = True
        self.event.save()
        self.project.set_public_access(self.role_guest)
        self.assert_response(self.url, self.anonymous, 403)


class TestEventStatusExtraAjaxViewSite(
    TimelineEventMixin, TimelineEventStatusMixin, ProjectPermissionTestBase
):
    """Tests for EventStatusExtraAjaxView permissions with a site event"""

    def setUp(self):
        super().setUp()
        self.event = self.make_event(
            None, APP_NAME_PR, self.user_owner, 'test_event'
        )
        self.event_status = self.make_event_status(
            self.event,
            'OK',
            extra_data={'example_data': 'example_extra_data'},
        )
        self.regular_user = self.make_user('regular_user')
        self.url = reverse(
            'timeline:ajax_extra_status',
            kwargs={
                'eventstatus': self.event_status.sodar_uuid,
            },
        )

    def test_get(self):
        """Test SiteEventExtraAjaxView GET"""
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, [self.regular_user, self.anonymous], 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(self.url, self.anonymous, 403)

    def test_get_classified(self):
        """Test GET with classified event"""
        self.event.classified = True
        self.event.save()
        self.assert_response(self.url, self.superuser, 200)
        self.assert_response(self.url, [self.regular_user, self.anonymous], 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_classified_anon(self):
        """Test GET with classified event and anonymous access"""
        self.event.classified = True
        self.event.save()
        self.assert_response(self.url, self.anonymous, 403)



from django_auth_ldap.backend import LDAPBackend, _LDAPUser
from django.conf import settings

# Username domains for primary and secondary LDAP backends
# Optional
LDAP_DOMAIN = getattr(settings, 'AUTH_LDAP_USERNAME_DOMAIN', None)
# Required for LDAP2
LDAP2_DOMAIN = getattr(settings, 'AUTH_LDAP2_USERNAME_DOMAIN', None)


# Primary LDAP backend
class PrimaryLDAPBackend(LDAPBackend):
    settings_prefix = 'AUTH_LDAP_'

    def authenticate(
        self, request=None, username=None, password=None, **kwargs
    ):
        # Login with username@DOMAIN
        if LDAP_DOMAIN:
            if (
                username.find('@') == -1
                or username.strip().split('@')[1].upper() != LDAP_DOMAIN
            ):
                return None
            ldap_user = _LDAPUser(self, username=username.split('@')[0].strip())
        # Login with username only
        else:
            if username.find('@') != -1:
                return None
            ldap_user = _LDAPUser(self, username=username.strip())
        user = ldap_user.authenticate(password)
        return user

    def ldap_to_django_username(self, username):
        """Override LDAPBackend function to get the username with domain"""
        return (username + '@' + LDAP_DOMAIN) if LDAP_DOMAIN else username

    def django_to_ldap_username(self, username):
        """Override LDAPBackend function to get the real LDAP username"""
        return username.split('@')[0] if LDAP_DOMAIN else username


# Secondary AD backend
class SecondaryLDAPBackend(LDAPBackend):
    settings_prefix = 'AUTH_LDAP2_'

    def authenticate(
        self, request=None, username=None, password=None, **kwargs
    ):
        if (
            username.find('@') == -1
            or username.split('@')[1].upper() != LDAP2_DOMAIN
        ):
            return None
        ldap_user = _LDAPUser(self, username=username.split('@')[0].strip())
        user = ldap_user.authenticate(password)
        return user

    def ldap_to_django_username(self, username):
        """Override LDAPBackend function to get the username with domain"""
        return username + '@' + LDAP2_DOMAIN

    def django_to_ldap_username(self, username):
        """Override LDAPBackend function to get the real LDAP username"""
        return username.split('@')[0]



"""Batchupdateroles management command for updating user roles in batch"""

import sys

from email.utils import parseaddr

from django.conf import settings
from django.contrib import auth
from django.core.management.base import BaseCommand
from django.http import HttpRequest
from django.utils import timezone

from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import (
    Project,
    Role,
    ProjectInvite,
    SODARUser,
    SODAR_CONSTANTS,
)
from projectroles.views import RoleAssignmentModifyMixin, ProjectInviteMixin
from projectroles.utils import build_secret


logger = ManagementCommandLogger(__name__)
User = auth.get_user_model()


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']


# HACK for issue #612
class MockRequest(HttpRequest):
    scheme = 'http'

    def mock_scheme(self, host):
        self.scheme = host.scheme

    def scheme(self):  # noqa
        return self.scheme


class Command(RoleAssignmentModifyMixin, ProjectInviteMixin, BaseCommand):
    help = 'Batch updates project roles and sends invites'

    roles = None
    owner_role = None
    del_role = None
    issuer = None
    update_count = 0
    invite_count = 0
    request = None

    # Internal helpers ---------------------------------------------------------

    def _make_request(self) -> HttpRequest:
        """
        Make HttpRequest to supply to view-based handlers.

        :return: HttpRequest object
        """
        host = settings.SODAR_API_DEFAULT_HOST
        request = MockRequest()
        request.mock_scheme(host)
        request.META['HTTP_HOST'] = host.hostname
        if host.port:
            request.META['HTTP_HOST'] += ':' + str(host.port)
        request.user = self.issuer
        return request

    def _update_role(self, project: Project, user: SODARUser, role: Role):
        """
        Handle role update for an existing user.

        :param project: Project object
        ;param user: SODARUser object
        :param role: Role object
        """
        logger.info(f'Updating role of user {user.username} to {role.name}..')
        role_as = project.get_role(user)
        if role_as and role_as.role == self.owner_role:
            logger.warning(
                'Skipping as ownership transfer is not permitted here'
            )
            return
        elif role_as and role_as.role == role:
            logger.info('Skipping as role already exists for user')
            return
        elif (
            role_as
            and role_as.project != project
            and role_as.role.rank > role.rank
        ):
            logger.warning(
                'Skipping as demoting an inherited role is not permitted'
            )
            return
        self.modify_assignment(
            data={'user': user, 'role': role},
            request=self.request,
            project=project,
            instance=role_as,
        )
        self.update_count += 1

    def _invite_user(self, email: str, project: Project, role: Role):
        """
        Create and send user for user not yet in system.

        :param email: String
        :param project: Project object
        :param role: Role object
        """
        logger.info(
            f'Creating and sending invite to {email} for role {role.name}..'
        )
        invite = ProjectInvite.objects.filter(
            email=email,
            project=project,
            active=True,
            date_expire__gte=timezone.now(),
        ).first()
        if invite:
            logger.info('Invite already exists for user in project')
            return
        invite = ProjectInvite.objects.create(
            email=email,
            project=project,
            role=role,
            issuer=self.issuer,
            secret=build_secret(),
        )
        self.handle_invite(invite, self.request, add_message=False)
        self.invite_count += 1

    def _handle_list_row(self, project: Project, role_name: str, email: str):
        """
        Handle row in CSV list.

        :param project: Project object
        :param role_name: String
        :param email: String
        """
        if role_name not in self.roles:
            raise Exception(f'Unknown role: "{role_name}"')
        elif role_name == SODAR_CONSTANTS['PROJECT_ROLE_OWNER']:
            raise Exception(
                'Ownership transfer not permitted in this operation, '
                'use appropriate UI or REST API endpoint'
            )
        role = self.roles[role_name]
        if not parseaddr(email)[1]:
            raise Exception(f'Invalid email: "{email}"')

        users = User.objects.filter(email=email)
        if users.count() > 1:
            logger.warning(
                f'Skipping due to multiple user accounts found for email '
                f'"{email}"'
            )
            return
        user = users.first()
        del_limit = getattr(settings, 'PROJECTROLES_DELEGATE_LIMIT', 1)

        if (
            role == self.del_role
            and del_limit != 0
            and len(project.get_delegates(inherited=False)) >= del_limit
        ):
            raise Exception(f'Delegate limit of {del_limit} has been reached')

        elif role == self.del_role and not self.issuer.has_perm(
            'projectroles.update_project_delegate', project
        ):
            raise Exception(
                'Issuer lacks perms to update delegates in this project'
            )

        # Update role for existing user in the system
        if user:
            try:
                self._update_role(project, user, role)
            except Exception as ex:
                raise Exception(f'Exception raised by _update_role(): {ex}')
        # Invite user not yet in the system
        else:
            try:
                self._invite_user(email, project, role)
            except Exception as ex:
                raise Exception(f'Exception raised by _invite_user(): {ex}')

    # Command ------------------------------------------------------------------

    def add_arguments(self, parser):
        parser.add_argument(
            '-f',
            '--file',
            dest='file',
            type=str,
            required=True,
            help='Semicolon separated file with project UUID, email and role '
            'columns',
        )
        parser.add_argument(
            '-i',
            '--issuer',
            dest='issuer',
            type=str,
            required=False,
            help='User name of issuing user (optional)',
        )

    def handle(self, *args, **options):
        logger.info('Batch updating project roles and/or sending invites..')
        logger.debug('File = {}'.format(options['file']))
        logger.debug('Issuer = {}'.format(options['issuer']))
        if (
            settings.PROJECTROLES_SITE_MODE == 'TARGET'
            and not settings.PROJECTROLES_TARGET_CREATE
        ):
            logger.error(
                'Operation not permitted on a target site without the ability '
                'to create local projects, aborting..'
            )
            sys.exit(1)

        if options['issuer']:
            self.issuer = User.objects.filter(
                username=options['issuer']
            ).first()
        else:
            self.issuer = User.objects.filter(
                username=settings.PROJECTROLES_DEFAULT_ADMIN
            ).first()

        if not self.issuer:
            logger.error(
                'Issuer not found with username "{}"'.format(options['issuer'])
            )
            sys.exit(1)

        file = options['file']
        try:
            with open(file, 'r') as f:
                file_data = [d for d in (line.strip() for line in f) if d]
        except Exception as ex:
            logger.error(f'Unable to read file: {ex}')
            sys.exit(1)

        # Validate file
        row_num = 1
        for d in file_data:
            ds = d.split(';')
            if len(ds) != 3 or len(ds[0]) != 36 or '@' not in ds[1]:
                logger.error(f'Invalid data in CSV file on row {row_num}')
                sys.exit(1)
            row_num += 1

        self.roles = {r.name: r for r in Role.objects.all()}
        self.owner_role = self.roles[SODAR_CONSTANTS['PROJECT_ROLE_OWNER']]
        self.del_role = self.roles[SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']]
        self.request = self._make_request()
        project_uuids = list(set([d.split(';')[0] for d in file_data]))
        error_count = 0

        for p_uuid in project_uuids:
            project = Project.objects.filter(sodar_uuid=p_uuid).first()
            if not project:
                logger.error(f'Project not found with UUID: {p_uuid}')
                continue
            if project.is_remote():
                logger.error(
                    f'Skipping remote {project.type.lower()} '
                    f'{project.get_log_title()}'
                )
                continue

            if not self.issuer.has_perm(
                'projectroles.update_project_members', project
            ) or not self.issuer.has_perm('projectroles.invite_users', project):
                logger.error(
                    f'Skipping project, issuer {self.issuer.username} lacks '
                    f'perms to update or invite members'
                )
                continue

            logger.info(
                f'Updating roles in {project.type.lower()} '
                f'{project.get_log_title()}..'
            )
            for ds in [
                d.split(';') for d in file_data if d.split(';')[0] == p_uuid
            ]:
                try:
                    self._handle_list_row(
                        project=project, role_name=ds[2].strip(), email=ds[1]
                    )
                except Exception as ex:
                    logger.error(ex)
                    error_count += 1
                    # if settings.DEBUG:
                    #     raise ex

        logger.info(
            'Update done: {} invite{} sent, {} role{} updated, '
            '{} error{}'.format(
                self.invite_count,
                's' if self.invite_count != 1 else '',
                self.update_count,
                's' if self.update_count != 1 else '',
                error_count,
                's' if error_count != 1 else '',
            )
        )



"""Plugins for the sodarcache app"""

# Projectroles dependency
from projectroles.plugins import BackendPluginPoint

from sodarcache.api import SODARCacheAPI


class BackendPlugin(BackendPluginPoint):
    """Plugin for registering backend app with Projectroles"""

    #: Name (used as plugin ID)
    name = 'sodar_cache'

    #: Title (used in templates)
    title = 'SODAR Cache Backend'

    #: Iconify icon
    icon = 'mdi:basket-outline'

    #: Description string
    description = (
        'SODAR Cache backend for caching and aggregating data from external '
        'databases'
    )

    def get_api(self, **kwargs) -> SODARCacheAPI:
        """Return API entry point object."""
        return SODARCacheAPI()



"""Tests for the project settings API in the projectroles app"""

import json

from django.test import override_settings

from test_plus.test import TestCase

from projectroles.app_settings import AppSettingAPI
from projectroles.models import Role, AppSetting, SODAR_CONSTANTS
from projectroles.plugins import PluginAppSettingDef, PluginAPI
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleAssignmentMixin,
    RemoteSiteMixin,
    RemoteProjectMixin,
    AppSettingMixin,
)


app_settings = AppSettingAPI()
plugin_api = PluginAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
REMOTE_LEVEL_READ_ROLES = SODAR_CONSTANTS['REMOTE_LEVEL_READ_ROLES']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_SCOPE_PROJECT_USER = SODAR_CONSTANTS[
    'APP_SETTING_SCOPE_PROJECT_USER'
]
APP_SETTING_SCOPE_SITE = SODAR_CONSTANTS['APP_SETTING_SCOPE_SITE']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']

# Local constants
APP_NAME = 'projectroles'
EXISTING_SETTING = 'project_bool_setting'
EXAMPLE_APP_NAME = 'example_project_app'
INVALID_SETTING_VALUE = 'INVALID VALUE'
INVALID_SETTING_MSG = 'INVALID_SETTING_VALUE detected'
S_PREFIX = 'settings.{}.'


class AppSettingInitMixin:
    """Mixing for initializing app setting data"""

    def init_app_settings(self):
        """Init test app settings"""
        # TODO: Rename these to match the settings in example_project_app
        # TODO: Only create these where it's necessary
        # Init test project settings
        self.project_str_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_str_setting',
            'setting_type': APP_SETTING_TYPE_STRING,
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.project_int_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_int_setting',
            'setting_type': APP_SETTING_TYPE_INTEGER,
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.project_str_setting_options = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_str_setting_options',
            'setting_type': APP_SETTING_TYPE_STRING,
            'value': 'string1',
            'options': ['string1', 'string2'],
            'update_value': 'string2',
            'non_valid_value': 'string3',
        }
        self.project_int_setting_options = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_int_setting_options',
            'setting_type': APP_SETTING_TYPE_INTEGER,
            'value': 0,
            'options': [0, 1],
            'update_value': 1,
            'non_valid_value': 2,
        }
        self.project_bool_settings = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_bool_setting',
            'setting_type': APP_SETTING_TYPE_BOOLEAN,
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.project_json_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'name': 'project_json_setting',
            'setting_type': APP_SETTING_TYPE_JSON,
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }
        # Init test user settings
        self.user_str_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_str_setting',
            'setting_type': APP_SETTING_TYPE_STRING,
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.user_int_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_int_setting',
            'setting_type': APP_SETTING_TYPE_INTEGER,
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.user_str_setting_options = {
            'plugin_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_str_setting_options',
            'setting_type': APP_SETTING_TYPE_STRING,
            'value': 'string1',
            'update_value': 'string2',
            'options': ['string1', 'string2'],
            'non_valid_value': False,
        }
        self.user_int_setting_options = {
            'plugin_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_int_setting_options',
            'setting_type': APP_SETTING_TYPE_INTEGER,
            'value': 0,
            'update_value': 1,
            'options': [0, 1],
            'non_valid_value': 'Nan',
        }
        self.user_bool_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_bool_setting',
            'setting_type': APP_SETTING_TYPE_BOOLEAN,
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.user_json_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'user': self.user,
            'name': 'user_json_setting',
            'setting_type': APP_SETTING_TYPE_JSON,
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }
        # Init test PROJECT_USER settings
        self.project_user_str_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_str_setting',
            'setting_type': APP_SETTING_TYPE_STRING,
            'value': 'test',
            'update_value': 'better test',
            'non_valid_value': False,
        }
        self.project_user_int_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_int_setting',
            'setting_type': APP_SETTING_TYPE_INTEGER,
            'value': 0,
            'update_value': 170,
            'non_valid_value': 'Nan',
        }
        self.project_user_bool_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_bool_setting',
            'setting_type': APP_SETTING_TYPE_BOOLEAN,
            'value': False,
            'update_value': True,
            'non_valid_value': 170,
        }
        self.project_user_json_setting = {
            'plugin_name': EXAMPLE_APP_NAME,
            'project': self.project,
            'user': self.user,
            'name': 'project_user_json_setting',
            'setting_type': APP_SETTING_TYPE_JSON,
            'value': {
                'Example': 'Value',
                'list': [1, 2, 3, 4, 5],
                'level_6': False,
            },
            'update_value': {'Test_more': 'often_always'},
            'non_valid_value': self.project,
        }

        self.settings = [
            self.project_int_setting,
            self.project_int_setting_options,
            self.project_json_setting,
            self.project_str_setting,
            self.project_str_setting_options,
            self.project_bool_settings,
            self.user_int_setting,
            self.user_int_setting_options,
            self.user_json_setting,
            self.user_str_setting,
            self.user_str_setting_options,
            self.user_bool_setting,
            self.project_user_int_setting,
            self.project_user_json_setting,
            self.project_user_str_setting,
            self.project_user_bool_setting,
        ]
        for s in self.settings:
            kwargs = {
                'plugin_name': s['plugin_name'],
                'name': s['name'],
                'setting_type': s['setting_type'],
                'value': (
                    s['value']
                    if s['setting_type'] != APP_SETTING_TYPE_JSON
                    else ''
                ),
                'value_json': (
                    s['value']
                    if s['setting_type'] == APP_SETTING_TYPE_JSON
                    else {}
                ),
            }
            if 'project' in s:
                kwargs['project'] = s['project']
            if 'user' in s:
                kwargs['user'] = s['user']
            self.make_setting(**kwargs)
        return self.settings


class TestAppSettingAPI(
    ProjectMixin,
    RoleAssignmentMixin,
    RemoteSiteMixin,
    RemoteProjectMixin,
    AppSettingMixin,
    AppSettingInitMixin,
    TestCase,
):
    """Tests for AppSettingAPI"""

    def setUp(self):
        # Init project
        self.project = self.make_project(
            title='TestProject', type=PROJECT_TYPE_PROJECT, parent=None
        )
        # Init role
        self.role_owner = Role.objects.get(name=PROJECT_ROLE_OWNER)
        # Init user & role
        self.user = self.make_user('owner')
        self.owner_as = self.make_assignment(
            self.project, self.user, self.role_owner
        )
        # Init test data
        self.settings = self.init_app_settings()
        example_plugin = plugin_api.get_app_plugin(EXAMPLE_APP_NAME)
        self.example_defs = example_plugin.app_settings

    def test_get(self):
        """Test get()"""
        for setting in self.settings:
            data = {
                'plugin_name': setting['plugin_name'],
                'setting_name': setting['name'],
            }
            if 'project' in setting:
                data['project'] = setting['project']
            if 'user' in setting:
                data['user'] = setting['user']
            val = app_settings.get(**data)
            self.assertEqual(val, setting['value'])

    def test_get_default(self):
        """Test get() with default value for existing setting"""
        s_defs = app_settings.get_definitions(
            scope=APP_SETTING_SCOPE_PROJECT, plugin_name=EXAMPLE_APP_NAME
        )
        default_val = s_defs[EXISTING_SETTING].default
        val = app_settings.get(
            plugin_name=EXAMPLE_APP_NAME,
            setting_name=EXISTING_SETTING,
            project=self.project,
        )
        self.assertEqual(val, default_val)

    def test_get_nonexisting(self):
        """Test get() with non-existing setting"""
        with self.assertRaises(KeyError):
            app_settings.get(
                plugin_name=EXAMPLE_APP_NAME,
                setting_name='NON-EXISTING SETTING',
                project=self.project,
            )

    def test_get_post_safe_json(self):
        """Test get() with JSON setting and post_safe=True"""
        val = app_settings.get(
            plugin_name=self.project_json_setting['plugin_name'],
            setting_name=self.project_json_setting['name'],
            project=self.project_json_setting['project'],
            post_safe=True,
        )
        self.assertEqual(type(val), str)

    def test_get_post_safe_str_none(self):
        """Test get() with string setting with null default and post_safe=True"""
        val = app_settings.get(
            plugin_name=self.project_str_setting['plugin_name'],
            setting_name=self.project_str_setting['name'],
            project=self.project_str_setting['project'],
            post_safe=True,
        )
        self.assertEqual(type(val), str)

    def test_get_all_by_scope_project(self):
        """Test get_all_by_scope() with PROJECT scope"""
        all_defs = app_settings.get_all_defs()
        vals = app_settings.get_all_by_scope(
            APP_SETTING_SCOPE_PROJECT, project=self.project
        )
        for k, v in all_defs.items():
            p_vals = [v for v in vals if v.startswith(S_PREFIX.format(k))]
            s_defs = [
                d for d in v.values() if d.scope == APP_SETTING_SCOPE_PROJECT
            ]
            self.assertEqual(len(p_vals), len(s_defs))
            for s_def in s_defs:
                self.assertEqual(
                    vals[f'settings.{k}.{s_def.name}'],
                    app_settings.get(
                        plugin_name=k,
                        setting_name=s_def.name,
                        project=self.project,
                    ),
                )

    def test_get_all_by_scope_user(self):
        """Test get_all_by_scope() with USER scope"""
        all_defs = app_settings.get_all_defs()
        vals = app_settings.get_all_by_scope(
            APP_SETTING_SCOPE_USER, user=self.user
        )
        for k, v in all_defs.items():
            p_vals = [v for v in vals if v.startswith(S_PREFIX.format(k))]
            s_defs = [
                d for d in v.values() if d.scope == APP_SETTING_SCOPE_USER
            ]
            self.assertEqual(len(p_vals), len(s_defs))
            for s_def in s_defs:
                self.assertEqual(
                    vals[f'settings.{k}.{s_def.name}'],
                    app_settings.get(
                        plugin_name=k,
                        setting_name=s_def.name,
                        user=self.user,
                    ),
                )

    def test_get_all_by_scope_project_user(self):
        """Test get_all_by_scope() with PROJECT_USER scope"""
        all_defs = app_settings.get_all_defs()
        vals = app_settings.get_all_by_scope(
            APP_SETTING_SCOPE_PROJECT_USER, project=self.project, user=self.user
        )
        for k, v in all_defs.items():
            p_vals = [v for v in vals if v.startswith(S_PREFIX.format(k))]
            s_defs = [
                d
                for d in v.values()
                if d.scope == APP_SETTING_SCOPE_PROJECT_USER
            ]
            self.assertEqual(len(p_vals), len(s_defs))
            for s_def in s_defs:
                self.assertEqual(
                    vals[f'settings.{k}.{s_def.name}'],
                    app_settings.get(
                        plugin_name=k,
                        setting_name=s_def.name,
                        project=self.project,
                        user=self.user,
                    ),
                )

    def test_get_all_by_scope_site(self):
        """Test get_all_by_scope() with SITE scope"""
        all_defs = app_settings.get_all_defs()
        vals = app_settings.get_all_by_scope(APP_SETTING_SCOPE_SITE)
        for k, v in all_defs.items():
            p_vals = [v for v in vals if v.startswith(S_PREFIX.format(k))]
            s_defs = [
                d for d in v.values() if d.scope == APP_SETTING_SCOPE_SITE
            ]
            self.assertEqual(len(p_vals), len(s_defs))
            for s_def in s_defs:
                self.assertEqual(
                    vals[f'settings.{k}.{s_def.name}'],
                    app_settings.get(
                        plugin_name=k,
                        setting_name=s_def.name,
                    ),
                )

    def test_get_all_by_scope_invalid_args(self):
        """Test get_all_by_scope() with invalid args"""
        with self.assertRaises(ValueError):
            app_settings.get_all_by_scope(
                APP_SETTING_SCOPE_PROJECT_USER, project=self.project  # No user
            )

    def test_is_set_no_object(self):
        """Test is_set() with no setting object set"""
        n = 'project_star'
        kw = {'project': self.project, 'user': self.user}
        self.assertFalse(
            AppSetting.objects.filter(app_plugin=None, name=n, **kw).exists()
        )
        self.assertFalse(
            app_settings.is_set(plugin_name=APP_NAME, setting_name=n, **kw)
        )

    def test_is_set_object_exists(self):
        """Test is_set() with existing setting object"""
        n = 'project_star'
        kw = {'project': self.project, 'user': self.user}
        app_settings.set(plugin_name=APP_NAME, setting_name=n, value=True, **kw)
        self.assertTrue(
            AppSetting.objects.filter(app_plugin=None, name=n, **kw).exists()
        )
        self.assertTrue(
            app_settings.is_set(plugin_name=APP_NAME, setting_name=n, **kw)
        )

    def test_is_set_default_value(self):
        """Test is_set() with existing setting object and default value"""
        n = 'project_star'
        kw = {'project': self.project, 'user': self.user}
        app_settings.set(
            plugin_name=APP_NAME, setting_name=n, value=False, **kw
        )
        self.assertTrue(
            AppSetting.objects.filter(app_plugin=None, name=n, **kw).exists()
        )
        self.assertTrue(
            app_settings.is_set(plugin_name=APP_NAME, setting_name=n, **kw)
        )

    def test_is_set_plugin(self):
        """Test is_set() with plugin setting"""
        n = 'project_str_setting'
        kw = {'project': self.project}
        # NOTE: Already created in setUp()
        self.assertTrue(
            AppSetting.objects.filter(
                app_plugin__name=EXAMPLE_APP_NAME, name=n, **kw
            ).exists()
        )
        self.assertTrue(
            app_settings.is_set(
                plugin_name=EXAMPLE_APP_NAME, setting_name=n, **kw
            )
        )

    def test_is_set_invalid_user(self):
        """Test is_set() with invalid user arg"""
        with self.assertRaises(ValueError):
            app_settings.is_set(
                plugin_name=APP_NAME,
                setting_name='project_star',
                project=self.project,
                user=None,
            )

    def test_set(self):
        """Test set()"""
        for setting in self.settings:
            data = {
                'plugin_name': setting['plugin_name'],
                'setting_name': setting['name'],
            }
            if 'project' in setting:
                data['project'] = setting['project']
            if 'user' in setting:
                data['user'] = setting['user']
            update_data = dict(data)
            update_data['value'] = setting['update_value']

            ret = app_settings.set(**update_data)
            self.assertEqual(ret, True)
            val = app_settings.get(**data)
            self.assertEqual(val, setting['update_value'])

    def test_set_unchanged(self):
        """Test set() with unchanged value"""
        for setting in self.settings:
            data = {
                'plugin_name': setting['plugin_name'],
                'setting_name': setting['name'],
            }
            if 'project' in setting:
                data['project'] = setting['project']
            if 'user' in setting:
                data['user'] = setting['user']

            update_data = dict(data)
            update_data['value'] = setting['value']

            ret = app_settings.set(**update_data)
            self.assertEqual(
                ret,
                False,
                msg='setting={}.{}'.format(
                    setting['plugin_name'], setting['name']
                ),
            )
            val = app_settings.get(**data)
            self.assertEqual(
                val,
                setting['value'],
                msg='setting={}.{}'.format(
                    setting['plugin_name'], setting['name']
                ),
            )

    def test_set_new(self):
        """Test set() with new but defined setting"""
        val = AppSetting.objects.get(
            app_plugin=plugin_api.get_app_plugin(EXAMPLE_APP_NAME).get_model(),
            project=self.project,
            name=EXISTING_SETTING,
        ).value
        self.assertEqual(bool(int(val)), False)

        ret = app_settings.set(
            plugin_name=EXAMPLE_APP_NAME,
            setting_name=EXISTING_SETTING,
            value=True,
            project=self.project,
        )
        self.assertEqual(ret, True)
        val = app_settings.get(
            plugin_name=EXAMPLE_APP_NAME,
            setting_name=EXISTING_SETTING,
            project=self.project,
        )
        self.assertEqual(True, val)
        setting = AppSetting.objects.get(
            app_plugin=plugin_api.get_app_plugin(EXAMPLE_APP_NAME).get_model(),
            project=self.project,
            name=EXISTING_SETTING,
        )
        self.assertIsInstance(setting, AppSetting)

    def test_set_undefined(self):
        """Test set() with undefined setting (should fail)"""
        with self.assertRaises(ValueError):
            app_settings.set(
                plugin_name=EXAMPLE_APP_NAME,
                setting_name='new_setting',
                value='new',
                project=self.project,
            )

    def test_set_multi_project_user(self):
        """Test set() with multiple instances of PROJECT_USER setting"""
        # Set up second user
        new_user = self.make_user('new_user')
        ret = app_settings.set(
            plugin_name=EXAMPLE_APP_NAME,
            setting_name='project_user_str_setting',
            project=self.project,
            user=self.user,
            value=True,
        )
        self.assertEqual(ret, True)
        ret = app_settings.set(
            plugin_name=EXAMPLE_APP_NAME,
            setting_name='project_user_str_setting',
            project=self.project,
            user=new_user,
            value=True,
        )
        self.assertEqual(ret, True)

    def test_set_invalid_project_types(self):
        """Test set() with invalid project types scope"""
        # Should fail because category_bool_setting has CATEGORY scope
        with self.assertRaises(ValueError):
            app_settings.set(
                plugin_name=EXAMPLE_APP_NAME,
                setting_name='category_bool_setting',
                project=self.project,
                value=True,
            )

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_set_target_local(self):
        """Test setting local setting on target site"""
        n = 'project_str_setting'
        v = 'updated'
        args = {'name': n, 'value': v}
        self.assertEqual(AppSetting.objects.filter(**args).count(), 0)
        app_settings.set(EXAMPLE_APP_NAME, n, v, project=self.project)
        self.assertEqual(AppSetting.objects.filter(**args).count(), 1)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_set_target_local_remote(self):
        """Test setting local setting on target site and remote project"""
        remote_site = self.make_site(
            name='Test source site',
            url='https://sodar.example.com',
            mode=SITE_MODE_SOURCE,
        )
        self.make_remote_project(
            project_uuid=self.project.sodar_uuid,
            site=remote_site,
            level=REMOTE_LEVEL_READ_ROLES,
            project=self.project,
        )
        n = 'project_str_setting'
        v = 'updated'
        args = {'name': n, 'value': v}
        self.assertEqual(AppSetting.objects.filter(**args).count(), 0)
        app_settings.set(EXAMPLE_APP_NAME, n, v, project=self.project)
        self.assertEqual(AppSetting.objects.filter(**args).count(), 1)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_set_target_global(self):
        """Test setting global setting on target site and local project"""
        n = 'project_global_setting'
        self.assertEqual(AppSetting.objects.filter(name=n, value=1).count(), 0)
        app_settings.set(EXAMPLE_APP_NAME, n, True, project=self.project)
        self.assertEqual(AppSetting.objects.filter(name=n, value=1).count(), 1)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_set_target_global_remote(self):
        """Test setting global setting on target site and remote project"""
        remote_site = self.make_site(
            name='Test source site',
            url='https://sodar.example.com',
            mode=SITE_MODE_SOURCE,
        )
        self.make_remote_project(
            project_uuid=self.project.sodar_uuid,
            site=remote_site,
            level=REMOTE_LEVEL_READ_ROLES,
            project=self.project,
        )
        n = 'project_global_setting'
        self.assertEqual(AppSetting.objects.filter(name=n, value=1).count(), 0)
        with self.assertRaises(ValueError):
            app_settings.set(EXAMPLE_APP_NAME, n, True, project=self.project)
        self.assertEqual(AppSetting.objects.filter(name=n, value=1).count(), 0)

    def test_set_user_global(self):
        """Test setting global user setting on source site"""
        n = 'notify_email_project'
        v = '0'
        self.assertEqual(AppSetting.objects.filter(name=n, value=v).count(), 0)
        app_settings.set(APP_NAME, n, False, user=self.user)
        self.assertEqual(AppSetting.objects.filter(name=n, value=v).count(), 1)

    @override_settings(PROJECTROLES_SITE_MODE=SITE_MODE_TARGET)
    def test_set_user_global_target(self):
        """Test setting global user setting on target site"""
        n = 'notify_email_project'
        v = '0'
        self.assertEqual(AppSetting.objects.filter(name=n, value=v).count(), 0)
        with self.assertRaises(ValueError):
            app_settings.set(APP_NAME, n, False, user=self.user)
        self.assertEqual(AppSetting.objects.filter(name=n, value=v).count(), 0)

    def test_validate_boolean(self):
        """Test validate() with type BOOLEAN"""
        for setting in self.settings:
            self.assertEqual(
                app_settings.validate(
                    setting['setting_type'],
                    setting['value'],
                    setting.get('options'),
                ),
                True,
            )
            if setting['setting_type'] == APP_SETTING_TYPE_STRING:
                continue
            with self.assertRaises(ValueError):
                app_settings.validate(
                    setting['setting_type'],
                    setting['non_valid_value'],
                    setting.get('options'),
                )

    def test_validate_int(self):
        """Test validate() with type INTEGER"""
        self.assertEqual(
            app_settings.validate(APP_SETTING_TYPE_INTEGER, 170, None), True
        )
        # NOTE: String is also OK if it corresponds to an int
        self.assertEqual(
            app_settings.validate(APP_SETTING_TYPE_INTEGER, '170', None), True
        )
        with self.assertRaises(ValueError):
            app_settings.validate(
                APP_SETTING_TYPE_INTEGER, 'not an integer', None
            )

    def test_validate_invalid(self):
        """Test validate() with invalid type"""
        with self.assertRaises(ValueError):
            app_settings.validate('INVALID_TYPE', 'value', None)

    def test_get_definition_plugin(self):
        """Test get_definition() with plugin"""
        app_plugin = plugin_api.get_app_plugin(EXAMPLE_APP_NAME)
        expected = {
            'name': 'project_str_setting',
            'scope': APP_SETTING_SCOPE_PROJECT,
            'type': APP_SETTING_TYPE_STRING,
            'label': 'String setting',
            'default': '',
            'description': 'Example string project setting',
            'placeholder': 'Example string',
            'user_modifiable': True,
            'global_edit': False,
            'options': [],
            'project_types': [PROJECT_TYPE_PROJECT],
            'widget_attrs': {},
        }
        s_def = app_settings.get_definition(
            'project_str_setting', plugin=app_plugin
        )
        self.assertIsInstance(s_def, PluginAppSettingDef)
        self.assertEqual(s_def.__dict__, expected)

    def test_get_definition_plugin_name(self):
        """Test get_definition() with plugin name"""
        expected = {
            'name': 'project_str_setting',
            'scope': APP_SETTING_SCOPE_PROJECT,
            'type': APP_SETTING_TYPE_STRING,
            'label': 'String setting',
            'default': '',
            'description': 'Example string project setting',
            'placeholder': 'Example string',
            'user_modifiable': True,
            'global_edit': False,
            'options': [],
            'project_types': [PROJECT_TYPE_PROJECT],
            'widget_attrs': {},
        }
        s_def = app_settings.get_definition(
            'project_str_setting', plugin_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(s_def.__dict__, expected)

    def test_get_definition_user(self):
        """Test get_definition() with user setting"""
        expected = {
            'name': 'user_str_setting',
            'scope': APP_SETTING_SCOPE_USER,
            'type': APP_SETTING_TYPE_STRING,
            'label': 'String setting',
            'default': '',
            'description': 'Example string user setting',
            'placeholder': 'Example string',
            'user_modifiable': True,
            'global_edit': False,
            'options': [],
            'project_types': [PROJECT_TYPE_PROJECT],
            'widget_attrs': {},
        }
        s_def = app_settings.get_definition(
            'user_str_setting', plugin_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(s_def.__dict__, expected)

    def test_get_definition_invalid(self):
        """Test get_definition() with innvalid input"""
        with self.assertRaises(ValueError):
            app_settings.get_definition(
                'non_existing_setting', plugin_name=EXAMPLE_APP_NAME
            )
        with self.assertRaises(ValueError):
            app_settings.get_definition(
                'project_str_setting', plugin_name='non_existing_app_name'
            )
        # Both app_name and plugin unset
        with self.assertRaises(ValueError):
            app_settings.get_definition('project_str_setting')

    def test_get_definitions_project(self):
        """Test get_definitions() with PROJECT scope"""
        defs = app_settings.get_definitions(
            APP_SETTING_SCOPE_PROJECT, plugin_name=EXAMPLE_APP_NAME
        )
        self.assertIsInstance(defs, dict)
        ex_defs = [
            d for d in self.example_defs if d.scope == APP_SETTING_SCOPE_PROJECT
        ]
        self.assertEqual(len(defs), len(ex_defs))
        self.assertTrue(
            all([d.scope == APP_SETTING_SCOPE_PROJECT for d in defs.values()])
        )

    def test_get_definitions_user(self):
        """Test get_definitions() with USER scope"""
        defs = app_settings.get_definitions(
            APP_SETTING_SCOPE_USER, plugin_name=EXAMPLE_APP_NAME
        )
        ex_defs = [
            d for d in self.example_defs if d.scope == APP_SETTING_SCOPE_USER
        ]
        self.assertEqual(len(defs), len(ex_defs))
        self.assertTrue(
            all([d.scope == APP_SETTING_SCOPE_USER for d in defs.values()])
        )

    def test_get_definitions_project_user(self):
        """Test get_definitions() with PROJECT_USER scope"""
        defs = app_settings.get_definitions(
            APP_SETTING_SCOPE_PROJECT_USER, plugin_name=EXAMPLE_APP_NAME
        )
        ex_defs = [
            d
            for d in self.example_defs
            if d.scope == APP_SETTING_SCOPE_PROJECT_USER
        ]
        self.assertEqual(len(defs), len(ex_defs))
        self.assertTrue(
            all(
                [
                    d.scope == APP_SETTING_SCOPE_PROJECT_USER
                    for d in defs.values()
                ]
            )
        )

    def test_get_definitions_site(self):
        """Test get_definitions() with SITE scope"""
        defs = app_settings.get_definitions(
            APP_SETTING_SCOPE_SITE, plugin_name=EXAMPLE_APP_NAME
        )
        ex_defs = [
            d for d in self.example_defs if d.scope == APP_SETTING_SCOPE_SITE
        ]
        self.assertEqual(len(defs), len(ex_defs))
        self.assertTrue(list(defs.values())[0].scope, APP_SETTING_SCOPE_SITE)

    def test_get_definitions_no_scope(self):
        """Test get_definitions() with no scope"""
        defs = app_settings.get_definitions(
            scope=None, plugin_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(len(defs), len(self.example_defs))

    def test_get_definitions_modifiable(self):
        """Test get_definitions() with user_modifiable arg"""
        defs = app_settings.get_definitions(
            APP_SETTING_SCOPE_PROJECT, plugin_name=EXAMPLE_APP_NAME
        )
        self.assertEqual(len(defs), 12)
        defs = app_settings.get_definitions(
            APP_SETTING_SCOPE_PROJECT,
            plugin_name=EXAMPLE_APP_NAME,
            user_modifiable=True,
        )
        self.assertEqual(len(defs), 10)

    def test_get_definitions_invalid_scope(self):
        """Test get_definitions() with invalid scope"""
        with self.assertRaises(ValueError):
            app_settings.get_definitions(
                'Ri4thai8aez5ooRa', plugin_name=EXAMPLE_APP_NAME
            )

    def test_get_defaults_project(self):
        """Test get_defaults() with PROJECT scope"""
        prefix = f'settings.{EXAMPLE_APP_NAME}.'
        defaults = app_settings.get_defaults(APP_SETTING_SCOPE_PROJECT)
        self.assertEqual(defaults[prefix + 'project_str_setting'], '')
        self.assertEqual(defaults[prefix + 'project_int_setting'], 0)
        self.assertEqual(defaults[prefix + 'project_bool_setting'], False)
        self.assertEqual(
            defaults[prefix + 'project_json_setting'],
            {'Example': 'Value', 'list': [1, 2, 3, 4, 5], 'level_6': False},
        )

    def test_get_defaults_project_post_safe(self):
        """Test get_defaults() with PROJECT scope and post_safe=True"""
        prefix = f'settings.{EXAMPLE_APP_NAME}.'
        defaults = app_settings.get_defaults(
            APP_SETTING_SCOPE_PROJECT, post_safe=True
        )
        self.assertEqual(defaults[prefix + 'project_str_setting'], '')
        self.assertEqual(defaults[prefix + 'project_int_setting'], 0)
        self.assertEqual(defaults[prefix + 'project_bool_setting'], False)
        self.assertEqual(
            defaults[prefix + 'project_json_setting'],
            json.dumps(
                {'Example': 'Value', 'list': [1, 2, 3, 4, 5], 'level_6': False}
            ),
        )

    def test_get_defaults_user(self):
        """Test get_defaults() with USER scope"""
        prefix = f'settings.{EXAMPLE_APP_NAME}.'
        defaults = app_settings.get_defaults(APP_SETTING_SCOPE_USER)
        self.assertEqual(defaults[prefix + 'user_str_setting'], '')
        self.assertEqual(defaults[prefix + 'user_int_setting'], 0)
        self.assertEqual(defaults[prefix + 'user_bool_setting'], False)
        self.assertEqual(
            defaults[prefix + 'user_json_setting'],
            {'Example': 'Value', 'list': [1, 2, 3, 4, 5], 'level_6': False},
        )

    def test_get_defaults_project_user(self):
        """Test get_defaults() with PROJECT_USER scope"""
        prefix = f'settings.{EXAMPLE_APP_NAME}.'
        defaults = app_settings.get_defaults(APP_SETTING_SCOPE_PROJECT_USER)
        self.assertEqual(defaults[prefix + 'project_user_str_setting'], '')
        self.assertEqual(defaults[prefix + 'project_user_int_setting'], 0)
        self.assertEqual(defaults[prefix + 'project_user_bool_setting'], False)
        self.assertEqual(
            defaults[prefix + 'project_user_json_setting'],
            {'Example': 'Value', 'list': [1, 2, 3, 4, 5], 'level_6': False},
        )

    def test_get_defaults_site(self):
        """Test get_defaults() with SITE scope"""
        prefix = f'settings.{EXAMPLE_APP_NAME}.'
        defaults = app_settings.get_defaults(APP_SETTING_SCOPE_SITE)
        self.assertEqual(defaults[prefix + 'site_bool_setting'], False)

    def test_delete_scope_project_params_none(self):
        """Test delete() with PROJECT scope and no params (should fail)"""
        self.assertEqual(AppSetting.objects.count(), 16)
        with self.assertRaises(ValueError):
            app_settings.delete(
                EXAMPLE_APP_NAME,
                'project_str_setting',
            )

    def test_delete_scope_project_params_user(self):
        """Test delete() with PROJECT scope and user param"""
        with self.assertRaises(ValueError):
            app_settings.delete(
                EXAMPLE_APP_NAME,
                'project_str_setting',
                user=self.user,
            )

    def test_delete_scope_project_params_project(self):
        """Test delete() with PROJECT scope and project param"""
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete(
            EXAMPLE_APP_NAME,
            'project_str_setting',
            project=self.project,
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_scope_project_params_user_project(self):
        """Test delete() with PROJECT scope and user/project params"""
        with self.assertRaises(ValueError):
            app_settings.delete(
                EXAMPLE_APP_NAME,
                'project_str_setting',
                project=self.project,
                user=self.user,
            )

    def test_delete_scope_user_params_none(self):
        """Test delete() with USER scope and no params (should fail)"""
        self.assertEqual(AppSetting.objects.count(), 16)
        with self.assertRaises(ValueError):
            app_settings.delete(EXAMPLE_APP_NAME, 'user_str_setting')

    def test_delete_scope_user_params_user(self):
        """Test delete() with USER scope and user param"""
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete(
            EXAMPLE_APP_NAME, 'user_str_setting', user=self.user
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_scope_user_params_project(self):
        """Test delete() with USER scope and project param"""
        with self.assertRaises(ValueError):
            app_settings.delete(
                EXAMPLE_APP_NAME, 'user_str_setting', project=self.project
            )

    def test_delete_scope_user_params_user_project(self):
        """Test delete() with USER scope and project/user params"""
        with self.assertRaises(ValueError):
            app_settings.delete(
                EXAMPLE_APP_NAME,
                'user_str_setting',
                project=self.project,
                user=self.user,
            )

    def test_delete_scope_project_user_params_none(self):
        """Test delete() with PROJECT_USER scope and no params (should fail)"""
        self.assertEqual(AppSetting.objects.count(), 16)
        with self.assertRaises(ValueError):
            app_settings.delete(
                EXAMPLE_APP_NAME,
                'project_user_str_setting',
                project=None,
                user=None,
            )
        self.assertEqual(AppSetting.objects.count(), 16)

    def test_delete_scope_project_user_params_user(self):
        """Test delete() with PROJECT_USER scope and user param (should fail)"""
        self.assertEqual(AppSetting.objects.count(), 16)
        with self.assertRaises(ValueError):
            app_settings.delete(
                EXAMPLE_APP_NAME,
                'project_user_str_setting',
                project=None,
                user=self.user,
            )
        self.assertEqual(AppSetting.objects.count(), 16)

    def test_delete_scope_project_user_params_project(self):
        """Test delete() with PROJECT_USER scope and project param"""
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete(
            EXAMPLE_APP_NAME,
            'project_user_str_setting',
            project=self.project,
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_scope_project_user_params_user_project(self):
        """Test delete() with PROJECT_USER scope and user/project params"""
        self.assertEqual(AppSetting.objects.count(), 16)
        app_settings.delete(
            EXAMPLE_APP_NAME,
            'project_user_str_setting',
            project=self.project,
            user=self.user,
        )
        self.assertEqual(AppSetting.objects.count(), 15)

    def test_delete_by_scope(self):
        """Test delete_by_scope()"""
        self.assertEqual(AppSetting.objects.count(), 16)
        # Delete PROJECT_USER scope settings
        app_settings.delete_by_scope(
            APP_SETTING_SCOPE_PROJECT_USER,
            project=self.project,
            user=self.user,
        )
        self.assertEqual(AppSetting.objects.count(), 12)
        # Delete PROJECT scope settings
        app_settings.delete_by_scope(
            APP_SETTING_SCOPE_USER,
            user=self.user,
        )
        self.assertEqual(AppSetting.objects.count(), 6)
        # Delete USER scope settings
        app_settings.delete_by_scope(
            APP_SETTING_SCOPE_PROJECT,
            project=self.project,
        )
        self.assertEqual(AppSetting.objects.count(), 0)

    def test_delete_by_scope_param_project(self):
        """Test delete_by_scope() with invalid project params"""
        with self.assertRaises(ValueError):
            app_settings.delete_by_scope(
                APP_SETTING_SCOPE_PROJECT,
                project=self.project,
                user=self.user,
            )

    def test_delete_by_scope_param_user(self):
        """Test delete_by_scope() with invalid user params"""
        with self.assertRaises(ValueError):
            app_settings.delete_by_scope(
                APP_SETTING_SCOPE_USER,
                project=self.project,
                user=self.user,
            )

    def test_validate_form_app_settings(self):
        """Test validate_form_app_settings() with valid project setting value"""
        app_plugin = plugin_api.get_app_plugin(EXAMPLE_APP_NAME)
        app_settings = {'project_str_setting': 'String'}
        errors = app_plugin.validate_form_app_settings(
            app_settings, project=self.project
        )
        self.assertEqual(errors, {})

    def test_validate_form_app_settings_invalid(self):
        """Test validate_form_app_settings() with invalid project setting value"""
        app_plugin = plugin_api.get_app_plugin(EXAMPLE_APP_NAME)
        app_settings = {'project_str_setting': INVALID_SETTING_VALUE}
        errors = app_plugin.validate_form_app_settings(
            app_settings, project=self.project
        )
        self.assertEqual(errors, {'project_str_setting': INVALID_SETTING_MSG})

    def test_validate_form_app_settings_user(self):
        """Test validate_form_app_settings() with valid user setting value"""
        app_plugin = plugin_api.get_app_plugin(EXAMPLE_APP_NAME)
        app_settings = {'user_str_setting': 'String'}
        errors = app_plugin.validate_form_app_settings(
            app_settings, user=self.user
        )
        self.assertEqual(errors, {})

    def test_validate_form_app_settings_user_invalid(self):
        """Test validate_form_app_settings() with invalid user setting value"""
        app_plugin = plugin_api.get_app_plugin(EXAMPLE_APP_NAME)
        app_settings = {'user_str_setting': INVALID_SETTING_VALUE}
        errors = app_plugin.validate_form_app_settings(
            app_settings, user=self.user
        )
        self.assertEqual(errors, {'user_str_setting': INVALID_SETTING_MSG})

    def test_compare_value_string(self):
        """Test compare_value() with string values"""
        n = 'project_str_setting'
        v = 'value'
        vf = 'valueXYZ'
        app_settings.set(EXAMPLE_APP_NAME, n, v, project=self.project)
        obj = AppSetting.objects.get(
            app_plugin__name=EXAMPLE_APP_NAME, name=n, project=self.project
        )
        self.assertEqual(app_settings.compare_value(obj, v), True)
        self.assertEqual(app_settings.compare_value(obj, vf), False)

    def test_compare_value_int(self):
        """Test compare_value() with int values"""
        n = 'project_int_setting'
        v = 0
        vf = 1
        app_settings.set(EXAMPLE_APP_NAME, n, v, project=self.project)
        obj = AppSetting.objects.get(
            app_plugin__name=EXAMPLE_APP_NAME, name=n, project=self.project
        )
        self.assertEqual(app_settings.compare_value(obj, v), True)
        self.assertEqual(app_settings.compare_value(obj, vf), False)
        self.assertEqual(app_settings.compare_value(obj, str(v)), True)
        self.assertEqual(app_settings.compare_value(obj, str(vf)), False)

    def test_compare_value_bool(self):
        """Test compare_value() with boolean values"""
        n = 'project_bool_setting'
        v = True
        vf = False
        app_settings.set(EXAMPLE_APP_NAME, n, v, project=self.project)
        obj = AppSetting.objects.get(
            app_plugin__name=EXAMPLE_APP_NAME, name=n, project=self.project
        )
        self.assertEqual(app_settings.compare_value(obj, v), True)
        self.assertEqual(app_settings.compare_value(obj, vf), False)
        self.assertEqual(app_settings.compare_value(obj, '1'), True)
        self.assertEqual(app_settings.compare_value(obj, '0'), False)

    def test_compare_value_json(self):
        """Test compare_value() with JSON values"""
        n = 'project_json_setting'
        v = {'x': 1, 'y': 2}
        vf = {'a': 3, 'b': 4}
        app_settings.set(EXAMPLE_APP_NAME, n, v, project=self.project)
        obj = AppSetting.objects.get(
            app_plugin__name=EXAMPLE_APP_NAME, name=n, project=self.project
        )
        self.assertEqual(app_settings.compare_value(obj, v), True)
        self.assertEqual(app_settings.compare_value(obj, vf), False)

    def test_compare_value_json_empty(self):
        """Test compare_value() with empty JSON values"""
        n = 'project_json_setting'
        v = {}
        vf = {'x': 1, 'y': 2}
        app_settings.set(EXAMPLE_APP_NAME, n, v, project=self.project)
        obj = AppSetting.objects.get(
            app_plugin__name=EXAMPLE_APP_NAME, name=n, project=self.project
        )
        self.assertEqual(app_settings.compare_value(obj, v), True)
        self.assertEqual(app_settings.compare_value(obj, vf), False)
        self.assertEqual(app_settings.compare_value(obj, None), True)
        self.assertEqual(app_settings.compare_value(obj, ''), True)



"""Factory Boy factory classes for the bgjobs app"""

import factory
from bgjobs.models import BackgroundJob


class BackgroundJobFactory(factory.django.DjangoModelFactory):
    """Factory for BackgroundJobFactory model."""

    class Meta:
        model = BackgroundJob

    # Can't set this because of circular dependency.
    # factory.SubFactory(ProjectFactory)
    project = None
    user = None  # Wait for SODAR core to offer a UserFactory
    job_type = 'type'



from django.urls import path

from . import views, views_api

# NOTE: file/folder/hyperlink objects can be referred to as 'item', but only if
#       ObjectPermissionMixin is used in the view

app_name = 'filesfolders'

urls_ui = [
    path(
        route='<uuid:project>',
        view=views.ProjectFileView.as_view(),
        name='list',
    ),
    path(
        route='folder/<uuid:folder>',
        view=views.ProjectFileView.as_view(),
        name='list',
    ),
    path(
        route='upload/<uuid:project>',
        view=views.FileCreateView.as_view(),
        name='file_create',
    ),
    path(
        route='upload/in/<uuid:folder>',
        view=views.FileCreateView.as_view(),
        name='file_create',
    ),
    path(
        route='update/<uuid:item>',
        view=views.FileUpdateView.as_view(),
        name='file_update',
    ),
    path(
        route='delete/<uuid:item>',
        view=views.FileDeleteView.as_view(),
        name='file_delete',
    ),
    path(
        route='download/<uuid:file>/<path:file_name>',
        view=views.FileServeView.as_view(),
        name='file_serve',
    ),
    path(
        route='download/<str:secret>/<path:file_name>',
        view=views.FileServePublicView.as_view(),
        name='file_serve_public',
    ),
    path(
        route='link/<uuid:file>',
        view=views.FilePublicLinkView.as_view(),
        name='file_public_link',
    ),
    path(
        route='folder/add/<uuid:project>',
        view=views.FolderCreateView.as_view(),
        name='folder_create',
    ),
    path(
        route='folder/add/in/<uuid:folder>',
        view=views.FolderCreateView.as_view(),
        name='folder_create',
    ),
    path(
        route='folder/update/<uuid:item>',
        view=views.FolderUpdateView.as_view(),
        name='folder_update',
    ),
    path(
        route='folder/delete/<uuid:item>',
        view=views.FolderDeleteView.as_view(),
        name='folder_delete',
    ),
    path(
        route='link/add/<uuid:project>',
        view=views.HyperLinkCreateView.as_view(),
        name='hyperlink_create',
    ),
    path(
        route='link/add/in/<uuid:folder>',
        view=views.HyperLinkCreateView.as_view(),
        name='hyperlink_create',
    ),
    path(
        route='link/update/<uuid:item>',
        view=views.HyperLinkUpdateView.as_view(),
        name='hyperlink_update',
    ),
    path(
        route='link/delete/<uuid:item>',
        view=views.HyperLinkDeleteView.as_view(),
        name='hyperlink_delete',
    ),
    path(
        route='batch/<uuid:project>',
        view=views.BatchEditView.as_view(),
        name='batch_edit',
    ),
    path(
        route='batch/in/<uuid:folder>',
        view=views.BatchEditView.as_view(),
        name='batch_edit',
    ),
]

urls_api = [
    path(
        route='api/folder/list-create/<uuid:project>',
        view=views_api.FolderListCreateAPIView.as_view(),
        name='api_folder_list_create',
    ),
    path(
        route='api/folder/retrieve-update-destroy/<uuid:folder>',
        view=views_api.FolderRetrieveUpdateDestroyAPIView.as_view(),
        name='api_folder_retrieve_update_destroy',
    ),
    path(
        route='api/file/list-create/<uuid:project>',
        view=views_api.FileListCreateAPIView.as_view(),
        name='api_file_list_create',
    ),
    path(
        route='api/file/retrieve-update-destroy/<uuid:file>',
        view=views_api.FileRetrieveUpdateDestroyAPIView.as_view(),
        name='api_file_retrieve_update_destroy',
    ),
    path(
        route='api/file/serve/<uuid:file>',
        view=views_api.FileServeAPIView.as_view(),
        name='api_file_serve',
    ),
    path(
        route='api/hyperlink/list-create/<uuid:project>',
        view=views_api.HyperLinkListCreateAPIView.as_view(),
        name='api_hyperlink_list_create',
    ),
    path(
        route='api/hyperlink/retrieve-update-destroy/<uuid:hyperlink>',
        view=views_api.HyperLinkRetrieveUpdateDestroyAPIView.as_view(),
        name='api_hyperlink_retrieve_update_destroy',
    ),
]

urlpatterns = urls_ui + urls_api



"""Tests for Django checks in the projectroles app"""

from django.test import override_settings

from test_plus.test import TestCase

import projectroles.checks as checks
from projectroles.apps import ProjectrolesConfig


# Local constants
AC = [ProjectrolesConfig]


class TestCheckAuthMethods(TestCase):
    """Tests for check_auth_methods()"""

    def test_check(self):
        """Test check_auth_methods() with default test settings"""
        self.assertEqual(checks.check_auth_methods(AC), [])

    @override_settings(PROJECTROLES_ALLOW_LOCAL_USERS=False)
    def test_check_disabled_auth(self):
        """Test check_auth_methods() with disabled auth settings"""
        # NOTE: Other settings are already false in the test config
        self.assertEqual(checks.check_auth_methods(AC), [checks.W001])



"""
Addremotesite management command for creating remote sites for remote project
sync
"""

import re
import sys

from django.conf import settings
from django.contrib import auth
from django.core.management.base import BaseCommand

from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import RemoteSite, SODAR_CONSTANTS
from projectroles.plugins import PluginAPI


logger = ManagementCommandLogger(__name__)
plugin_api = PluginAPI()
User = auth.get_user_model()


# SODAR constants
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_PEER = SODAR_CONSTANTS['SITE_MODE_PEER']


class Command(BaseCommand):
    help = 'Creates a remote site from given arguments'

    def add_arguments(self, parser):
        # RemoteSite Model Argumments
        parser.add_argument(
            '-n',
            '--name',
            dest='name',
            type=str,
            required=True,
            help='Name of the remote site',
        )
        parser.add_argument(
            '-u',
            '--url',
            dest='url',
            type=str,
            required=True,
            help='URL of the remote site. Can be provided without protocol '
            'prefix, defaults to HTTP',
        )
        parser.add_argument(
            '-m',
            '--mode',
            dest='mode',
            type=str,
            required=True,
            help='Mode of the remote site',
        )
        parser.add_argument(
            '-d',
            '--description',
            dest='description',
            type=str,
            required=False,
            default='',
            help='Description of the remote site',
        )
        parser.add_argument(
            '-t',
            '--token',
            dest='secret',
            type=str,
            required=True,
            help='Secret token of the remote site',
        )
        parser.add_argument(
            '-ud',
            '--user-display',
            dest='user_display',
            default=True,
            required=False,
            type=bool,
            help='User display of the remote site',
        )
        parser.add_argument(
            '-o',
            '--owner-modifiable',
            dest='owner_modifiable',
            default=True,
            required=False,
            type=bool,
            help='Allow owners and delegates to modify project access for this '
            'site',
        )
        # Additional Arguments
        parser.add_argument(
            '-s',
            '--suppress-error',
            dest='suppress_error',
            required=False,
            default=False,
            action='store_true',
            help='Suppresses error if site already exists',
        )

    def handle(self, *args, **options):
        timeline = plugin_api.get_backend_api('timeline_backend')
        logger.info('Creating remote site..')
        name = options['name']
        url = options['url']

        # Validate URL
        if not url.startswith('http://') and not url.startswith('https://'):
            url = ''.join(['http://', url])
        pattern = re.compile(r'(http|https)://.*\..*')
        if not pattern.match(url):
            logger.error(f'Invalid URL "{url}"')
            sys.exit(1)

        # Validate site mode
        site_mode = options['mode'].upper()
        host_mode = settings.PROJECTROLES_SITE_MODE
        if site_mode not in [SITE_MODE_SOURCE, SITE_MODE_TARGET]:
            logger.error(f'Invalid mode "{site_mode}"')
            sys.exit(1)
        if site_mode == host_mode:
            logger.error('Attempting to create site with the same mode as host')
            sys.exit(1)

        # Validate whether site exists
        name_exists = RemoteSite.objects.filter(name=name).count()
        url_exists = RemoteSite.objects.filter(url=url).count()
        if name_exists or url_exists:
            err_msg = 'Remote site exists with {} "{}"'.format(
                'name' if name_exists else 'URL', name if name_exists else url
            )
            if not options['suppress_error']:
                logger.error(err_msg)
                sys.exit(1)
            else:
                logger.info(err_msg)
                sys.exit(0)

        create_kw = {
            'name': name,
            'url': url,
            'mode': site_mode,
            'description': options['description'],
            'secret': options['secret'],
            'user_display': options['user_display'],
            'owner_modifiable': options['owner_modifiable'],
        }
        site = RemoteSite.objects.create(**create_kw)

        if timeline:
            tl_event = timeline.add_event(
                project=None,
                app_name='projectroles',
                user=None,
                event_name=f'{site_mode.lower()}_site_create',
                description='create {} remote site {{{}}} via management '
                'command'.format(site_mode.lower(), 'remote_site'),
                classified=True,
                status_type=timeline.TL_STATUS_OK,
                extra_data=create_kw,
            )
            tl_event.add_object(obj=site, label='remote_site', name=site.name)
        logger.info(f'Created remote site "{site.name}" with mode {site.mode}')



"""Tests for the API in the sodarcache app"""

from django.forms.models import model_to_dict

# Projectroles dependency
from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAPI


from sodarcache.models import JSONCacheItem
from sodarcache.tests.test_models import (
    JSONCacheItemTestBase,
    JSONCacheItemMixin,
)


plugin_api = PluginAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME = 'sodarcache'
INVALID_APP_NAME = 'timeline'  # Valid app name but but no data is created


class TestSodarCacheAPI(JSONCacheItemMixin, JSONCacheItemTestBase):
    """Testing sodarcache API class"""

    def setUp(self):
        super().setUp()
        self.cache_backend = plugin_api.get_backend_api('sodar_cache')

    def test_get_model(self):
        """Test get_model()"""
        self.assertEqual(self.cache_backend.get_model(), JSONCacheItem)

    def test_set_cache_item(self):
        """Test set_cache_item() to create item"""
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)
        item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        expected = {
            'id': item.pk,
            'project': self.project.pk,
            'app_name': APP_NAME,
            'user': self.user_owner.pk,
            'name': 'test_item',
            'data': {'test_key': 'test_val'},
            'sodar_uuid': item.sodar_uuid,
        }
        self.assertEqual(model_to_dict(item), expected)

    def test_set_cache_item_invalid_app_name(self):
        """Test set_cache_item() with invalid app name"""
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)
        with self.assertRaises(ValueError):
            self.cache_backend.set_cache_item(
                project=self.project,
                app_name='NON-EXISTING APP NAME',
                user=self.user_owner,
                name='test_item',
                data={'test_key': 'test_val'},
            )
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)

    def test_set_cache_item_invalid_data_type(self):
        """Test set_cache_item() with invalid data type"""
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)
        with self.assertRaises(ValueError):
            self.cache_backend.set_cache_item(
                project=self.project,
                app_name='NON-EXISTING APP NAME',
                user=self.user_owner,
                name='test_item',
                data_type='INVALID DATA TYPE',
                data={'test_key': 'test_val'},
            )
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)

    def test_set_cache_item_update(self):
        """Test set_cache_item() to update existing event"""
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)
        item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        update_item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'new_test_val'},
        )
        expected = {
            'id': item.pk,
            'project': self.project.pk,
            'app_name': APP_NAME,
            'user': self.user_owner.pk,
            'name': 'test_item',
            'data': {'test_key': 'new_test_val'},
            'sodar_uuid': item.sodar_uuid,
        }
        self.assertEqual(model_to_dict(update_item), expected)

    def test_set_cache_item_update_no_user(self):
        """Test set_cache_item() to update item with no user"""
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)
        item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        update_item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            name='test_item',
            data={'test_key': 'new_test_val'},
        )
        expected = {
            'id': item.pk,
            'project': self.project.pk,
            'app_name': APP_NAME,
            'user': None,
            'name': 'test_item',
            'data': {'test_key': 'new_test_val'},
            'sodar_uuid': item.sodar_uuid,
        }
        self.assertEqual(model_to_dict(update_item), expected)

    def test_get_cache_item(self):
        """Test get_cache_item()"""
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)
        item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        get_item = self.cache_backend.get_cache_item(
            app_name=APP_NAME, name='test_item', project=self.project
        )
        expected = {
            'id': item.pk,
            'project': self.project.pk,
            'app_name': APP_NAME,
            'user': self.user_owner.pk,
            'name': 'test_item',
            'data': {'test_key': 'test_val'},
            'sodar_uuid': item.sodar_uuid,
        }
        self.assertEqual(model_to_dict(get_item), expected)

    def test_delete_cache_item(self):
        """Test delete_cache_item()"""
        self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        item = self.cache_backend.get_cache_item(
            app_name=APP_NAME, name='test_item', project=self.project
        )
        self.assertIsNotNone(item)
        self.cache_backend.delete_cache_item(
            app_name=APP_NAME, name='test_item', project=self.project
        )
        item = self.cache_backend.get_cache_item(
            app_name=APP_NAME, name='test_item', project=self.project
        )
        self.assertIsNone(item)
        # Test for deleting a non-existing item
        self.cache_backend.delete_cache_item(
            app_name=APP_NAME, name='test_item', project=self.project
        )

    def test_get_project_cache(self):
        """Test get_project_cache()"""
        first_item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item1',
            data={'test_key1': 'test_val1'},
        )
        second_item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item2',
            data={'test_key2': 'test_val2'},
        )
        project_items = self.cache_backend.get_project_cache(
            project=self.project, data_type='json'
        )
        self.assertEqual(project_items.count(), 2)
        self.assertIn(first_item, project_items)
        self.assertIn(second_item, project_items)

    def test_get_update_time(self):
        """Test get_update_time()"""
        item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        update_time = self.cache_backend.get_update_time(
            app_name=APP_NAME, name='test_item', project=self.project
        )
        self.assertEqual(update_time, item.date_modified.timestamp())

    def test_delete_cache(self):
        """Test delete_cache()"""
        self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        delete_status = self.cache_backend.delete_cache()
        self.assertEqual(delete_status, 1)
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)

    def test_delete_cache_app_name(self):
        """Test delete_cache() with app name"""
        self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        delete_status = self.cache_backend.delete_cache(app_name=APP_NAME)
        self.assertEqual(delete_status, 1)
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)

    def test_delete_cache_app_name_no_items(self):
        """Test delete_cache() with app name and no created items"""
        self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        delete_status = self.cache_backend.delete_cache(
            app_name=INVALID_APP_NAME
        )
        self.assertEqual(delete_status, 0)
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)

    def test_delete_cache_project(self):
        """Test delete_cache() with project"""
        self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        delete_status = self.cache_backend.delete_cache(project=self.project)
        self.assertEqual(delete_status, 1)
        self.assertEqual(JSONCacheItem.objects.all().count(), 0)

    def test_delete_cache_project_no_items(self):
        """Test delete_cache() with project and no created items"""
        new_project = self.make_project(
            'NewProject', PROJECT_TYPE_PROJECT, None
        )
        self.cache_backend.set_cache_item(
            project=self.project,
            app_name=APP_NAME,
            user=self.user_owner,
            name='test_item',
            data={'test_key': 'test_val'},
        )
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)
        delete_status = self.cache_backend.delete_cache(project=new_project)
        self.assertEqual(delete_status, 0)
        self.assertEqual(JSONCacheItem.objects.all().count(), 1)



"""Template tags for role management views in projectroles"""

from typing import Optional

from django import template

from projectroles.models import (
    Project,
    SODARUser,
    Role,
    RoleAssignment,
    SODAR_CONSTANTS,
)


register = template.Library()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']

# Local constants
ROLE_RANK_ICONS = {
    10: 'mdi:star',
    20: 'mdi:star-half-full',
    30: 'mdi:account',
}


@register.simple_tag
def get_role_icon(role: Role) -> str:
    """Return the icon class for a Role"""
    keys = list(ROLE_RANK_ICONS.keys())
    key_idx = min(range(len(keys)), key=lambda i: abs(keys[i] - role.rank))
    return ROLE_RANK_ICONS[keys[key_idx]]


@register.simple_tag
def get_role_class(user: SODARUser) -> str:
    """
    Return class string for user role based on user status.

    :param user: SODARUser object
    :return: String
    """
    if not user.is_active:
        return 'text-secondary'
    return ''


@register.simple_tag
def get_role_perms(project: Project, user: SODARUser) -> dict:
    """
    Return role perms to template as dict.

    :param project: Project object
    :param user: SODARUser object
    :return: Dict
    """
    return {
        'can_update_owner': user.has_perm(
            'projectroles.update_project_owner', project
        ),
        'can_update_delegate': user.has_perm(
            'projectroles.update_project_delegate', project
        ),
        'can_update_members': user.has_perm(
            'projectroles.update_project_members', project
        ),
        'can_invite': user.has_perm('projectroles.invite_users', project),
    }


@register.simple_tag
def display_role_buttons(
    project: Project, role_as: RoleAssignment, perms: dict
) -> bool:
    """Return True/False on whether role buttons can be displayed for role"""
    if (
        role_as.role.name == PROJECT_ROLE_OWNER
        and role_as.project == project
        and perms['can_update_owner']
    ):
        return True
    if (
        role_as.role.name == PROJECT_ROLE_DELEGATE
        and role_as.project == project
        and perms['can_update_delegate']
    ):
        return True
    if (
        role_as.role.name not in [PROJECT_ROLE_OWNER, PROJECT_ROLE_DELEGATE]
        and perms['can_update_members']
    ):
        return True
    return False


@register.simple_tag
def get_inactive_role(
    project: Project, inherited_as: RoleAssignment
) -> Optional[RoleAssignment]:
    """
    Return inactive role assignment for inherited user if it exists.

    :param project: Project object
    :param inherited_as: RoleAssignment for inherited role
    :return: RoleAssignment object or None
    """
    local_as = RoleAssignment.objects.filter(
        project=project, user=inherited_as.user
    ).first()
    if local_as and local_as.role.rank > inherited_as.role.rank:
        return local_as
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=C0111,W6005,W6100
"""
Package metadata for django-sodar-core.
"""
import os
from setuptools import setup

import versioneer

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.

    Returns:
        list: Requirements file relative path strings
    """
    requirements = set()
    for path in requirements_paths:
        requirements.update(
            line.split('#')[0].strip()
            for line in open(path).readlines()
            if is_requirement(line.strip())
        )
    return list(requirements)


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.

    Returns:
        bool: True if the line is not blank, a comment, a URL, or an included
              file
    """
    return not (
        line == ''
        or line.startswith('-r')
        or line.startswith('#')
        or line.startswith('-e')
        or line.startswith('git+')
    )


README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

setup(
    name='django-sodar-core',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="""SODAR Core framework and project management apps""",
    long_description=README,
    long_description_content_type='text/x-rst',
    author='Berlin Institute of Health, Core Unit Bioinformatics',
    author_email='cubi@bih-charite.de',
    url='https://github.com/bihealth/sodar-core',
    packages=[
        'projectroles',
        'adminalerts',
        'appalerts',
        'bgjobs',
        'filesfolders',
        'siteinfo',
        'sodarcache',
        'timeline',
        'tokens',
        'userprofile',
    ],
    include_package_data=True,
    install_requires=load_requirements('requirements/base.txt'),
    zip_safe=False,
    classifiers=[
        'Framework :: Django',
        'Framework :: Django :: 4.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    python_requires='>=3.9',
)



from django.urls import path

from userprofile import views

app_name = 'userprofile'

urlpatterns = [
    path(
        route='profile',
        view=views.UserDetailView.as_view(),
        name='detail',
    ),
    path(
        route='settings/update',
        view=views.UserAppSettingsView.as_view(),
        name='settings_update',
    ),
    path(
        route='email/create',
        view=views.UserEmailCreateView.as_view(),
        name='email_create',
    ),
    path(
        route='email/verify/<str:secret>',
        view=views.UserEmailVerifyView.as_view(),
        name='email_verify',
    ),
    path(
        route='email/resend/<uuid:sodaruseradditionalemail>',
        view=views.UserEmailVerifyResendView.as_view(),
        name='email_verify_resend',
    ),
    path(
        route='email/delete/<uuid:sodaruseradditionalemail>',
        view=views.UserEmailDeleteView.as_view(),
        name='email_delete',
    ),
]



"""Projectroles app settings API"""

import json
import logging

from typing import Any, Optional, Union

from django.conf import settings

from djangoplugins.point import PluginPoint

from projectroles.models import AppSetting, Project, SODARUser, SODAR_CONSTANTS
from projectroles.plugins import PluginAppSettingDef, PluginAPI
from projectroles.utils import get_display_name


logger = logging.getLogger(__name__)
plugin_api = PluginAPI()


# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
SITE_MODE_SOURCE = SODAR_CONSTANTS['SITE_MODE_SOURCE']
SITE_MODE_TARGET = SODAR_CONSTANTS['SITE_MODE_TARGET']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_SCOPE_PROJECT_USER = SODAR_CONSTANTS[
    'APP_SETTING_SCOPE_PROJECT_USER'
]
APP_SETTING_SCOPE_SITE = SODAR_CONSTANTS['APP_SETTING_SCOPE_SITE']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']

# Local constants
APP_NAME = 'projectroles'
APP_SETTING_GLOBAL_DEFAULT = False
APP_SETTING_DEFAULT_VALUES = {
    APP_SETTING_TYPE_BOOLEAN: False,
    APP_SETTING_TYPE_INTEGER: 0,
    APP_SETTING_TYPE_JSON: {},
    APP_SETTING_TYPE_STRING: '',
}
APP_SETTING_SCOPE_ARGS = {
    APP_SETTING_SCOPE_PROJECT: {'project': True, 'user': False},
    APP_SETTING_SCOPE_USER: {'project': False, 'user': True},
    APP_SETTING_SCOPE_PROJECT_USER: {'project': True, 'user': True},
    APP_SETTING_SCOPE_SITE: {'project': False, 'user': False},
}
PROJECT_LIST_PAGE_OPTIONS = [10, 25, 50, 100, (-1, 'No pagination')]
DELETE_SCOPE_ERR_MSG = 'Argument "{arg}" must be set for {scope} scope setting'
GLOBAL_PROJECT_ERR_MSG = (
    'Overriding global settings for remote projects not allowed'
)
GLOBAL_USER_ERR_MSG = (
    'Overriding global user settings on target site not allowed'
)


# Define App Settings for projectroles app
PROJECTROLES_APP_SETTINGS = [
    PluginAppSettingDef(
        name='ip_restrict',
        scope=APP_SETTING_SCOPE_PROJECT,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=False,
        label='IP restrict',
        description='Restrict project access to the IP allow list',
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='ip_allow_list',
        scope=APP_SETTING_SCOPE_PROJECT,
        type=APP_SETTING_TYPE_STRING,
        label='IP allow list',
        description='Comma-separated list of allowed IPs for project access',
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='project_star',
        scope=APP_SETTING_SCOPE_PROJECT_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=False,
        global_edit=False,
        project_types=[PROJECT_TYPE_PROJECT, PROJECT_TYPE_CATEGORY],
    ),
    PluginAppSettingDef(
        name='notify_alert_project',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=True,
        label='Receive alerts for {} updates'.format(
            get_display_name(PROJECT_TYPE_PROJECT)
        ),
        description=(
            'Receive UI alerts for {} or {} creation, updating, moving, '
            'archiving and deletion.'.format(
                get_display_name(PROJECT_TYPE_CATEGORY),
                get_display_name(PROJECT_TYPE_PROJECT),
            )
        ),
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='notify_email_project',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=True,
        label='Receive email for {} updates'.format(
            get_display_name(PROJECT_TYPE_PROJECT)
        ),
        description=(
            'Receive email notifications for {} or {} creation, updating, '
            'moving, archiving and deletion.'.format(
                get_display_name(PROJECT_TYPE_CATEGORY),
                get_display_name(PROJECT_TYPE_PROJECT),
            )
        ),
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='notify_alert_role',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=True,
        label='Receive alerts for {} membership updates'.format(
            get_display_name(PROJECT_TYPE_PROJECT)
        ),
        description=(
            'Receive UI alerts for {} or {} membership updates and invitation '
            'activity.'.format(
                get_display_name(PROJECT_TYPE_CATEGORY),
                get_display_name(PROJECT_TYPE_PROJECT),
            )
        ),
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='notify_email_role',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=True,
        label='Receive email for {} membership updates'.format(
            get_display_name(PROJECT_TYPE_PROJECT)
        ),
        description=(
            'Receive email notifications for {} or {} membership updates and '
            'invitation activity.'.format(
                get_display_name(PROJECT_TYPE_CATEGORY),
                get_display_name(PROJECT_TYPE_PROJECT),
            )
        ),
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='site_read_only',
        scope=APP_SETTING_SCOPE_SITE,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=False,
        label='Site read-only mode',
        description='Set site in read-only mode. Data altering operations will '
        'be prohibited. Mode must be explicitly unset to allow data '
        'modification.',
        user_modifiable=True,
        global_edit=False,
    ),
    PluginAppSettingDef(
        name='project_list_highlight',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=True,
        label='Project list title highlight',
        description='Highlight project title in paths displayed in the project '
        'list.',
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='project_list_home_starred',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=False,
        label='Home view project list starred display',
        description='Initial starred mode to display in home view project list',
        user_modifiable=False,  # Editable only in project list
        global_edit=False,
    ),
    PluginAppSettingDef(
        name='project_list_pagination',
        scope=APP_SETTING_SCOPE_USER,
        type=APP_SETTING_TYPE_INTEGER,
        default=10,
        label='Project list page size',
        description='Amount of projects per page in the project list.',
        options=PROJECT_LIST_PAGE_OPTIONS,
        user_modifiable=True,
        global_edit=True,
    ),
    PluginAppSettingDef(
        name='project_access_block',
        scope=APP_SETTING_SCOPE_PROJECT,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=False,
        label='Block project access',
        description='Temporarily block all user access to project.',
        user_modifiable=False,  # Only allow modification by superuser
        global_edit=False,
        project_types=[PROJECT_TYPE_PROJECT],
    ),
    PluginAppSettingDef(
        name='category_public_stats',
        scope=APP_SETTING_SCOPE_PROJECT,
        type=APP_SETTING_TYPE_BOOLEAN,
        default=False,
        label='Display public category statistics',
        description='Enable all users to view {} readme and statistics. '
        'Limited to top level {}.'.format(
            get_display_name(PROJECT_TYPE_CATEGORY),
            get_display_name(PROJECT_TYPE_CATEGORY, plural=True),
        ),
        user_modifiable=True,  # Only allow modification by superuser
        global_edit=False,
        project_types=[PROJECT_TYPE_CATEGORY],
    ),
]


class AppSettingAPI:
    @classmethod
    def _validate_project_and_user(
        cls, scope: str, project: Project, user: SODARUser
    ):
        """
        Ensure project and user parameters are set according to scope.

        :param scope: Scope of Setting (USER, PROJECT, PROJECT_USER, SITE)
        :param project: Project object
        :param user: User object
        :raise: ValueError if none or both objects exist
        """
        if not APP_SETTING_SCOPE_ARGS[scope] == {
            'project': project is not None,
            'user': user is not None,
        }:
            raise ValueError(
                'Project and/or user invalid for setting with {} scope '
                '(project={}, user={})'.format(
                    scope,
                    project.get_log_title() if project else None,
                    user.username if user else None,
                )
            )

    @classmethod
    def _validate_value_in_options(
        cls,
        setting_value: str,
        setting_options: Union[list, callable],
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Ensure setting_value is present in setting_options.

        :param setting_value: String
        :param setting_options: List of options
        :param project: Project object
        :param user: User object
        :raise: ValueError if type is not recognized
        """
        if callable(setting_options):
            valid_options = [
                val[0] if isinstance(val, tuple) else val
                for val in setting_options(project, user)
            ]
            if setting_value not in valid_options:
                raise ValueError(
                    'Choice "{}" not found in options ({})'.format(
                        setting_value,
                        ', '.join(map(str, valid_options)),
                    )
                )
        elif setting_options:
            opts = [
                o[0] if isinstance(o, tuple) else o for o in setting_options
            ]
            if setting_value not in opts:
                raise ValueError(
                    'Choice "{}" not found in options ({})'.format(
                        setting_value, ', '.join(map(str, setting_options))
                    )
                )

    @classmethod
    def _get_app_plugin(cls, plugin_name: str) -> PluginPoint:
        """
        Return app plugin by name.

        :param plugin_name: Name of the app plugin (string)
        :return: App plugin object
        :raise: ValueError if plugin is not found with the name
        """
        plugin = plugin_api.get_app_plugin(plugin_name)
        if not plugin:
            raise ValueError(f'Plugin not found with name "{plugin_name}"')
        return plugin

    @classmethod
    def _get_defs(
        cls,
        plugin: Optional[PluginPoint] = None,
        plugin_name: Optional[str] = None,
    ) -> dict:
        """
        Get app setting definitions for a plugin.

        :param plugin: Plugin object or None
        :param plugin_name: Name of the app plugin (string or None)
        :return: Dict
        :raise: ValueError if args are not valid or plugin is not found
        """
        if not plugin and not plugin_name:
            raise ValueError('Plugin object and name both unset')
        if plugin_name == APP_NAME:
            return cls.get_projectroles_defs()
        if not plugin:
            plugin = cls._get_app_plugin(plugin_name)
        return {s.name: s for s in plugin.app_settings}

    @classmethod
    def _get_json_value(cls, value: Union[str, dict]) -> dict:
        """
        Return JSON value as dict regardless of input type

        :param value: Original value (string or dict)
        :raise: json.decoder.JSONDecodeError if string value is not valid JSON
        :raise: ValueError if value type is not recognized or if value is not
                valid JSON
        :return: dict
        """
        if not value:
            return {}
        try:
            if isinstance(value, str):
                return json.loads(value)
            json.dumps(value)  # Ensure this is valid
            return value
        except Exception:
            raise ValueError(f'Value is not valid JSON: {value}')

    @classmethod
    def _log_set_debug(
        cls,
        action: str,
        plugin_name: str,
        setting_name: str,
        value: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Helper method for logging setting changes in set() method.

        :param action: Action string (string)
        :param plugin_name: App plugin name (string)
        :param setting_name: Setting name (string)
        :param value: Setting value (string)
        :param project: Project object or None
        :param user: User object or None
        """
        extra_data = []
        if project:
            extra_data.append(f'project={project.sodar_uuid}')
        if user:
            extra_data.append(f'user={user.username}')
        logger.debug(
            '{} app setting: {}.{} = "{}"{}'.format(
                action,
                plugin_name,
                setting_name,
                value,
                ' ({})'.format('; '.join(extra_data)) if extra_data else '',
            )
        )

    @classmethod
    def get_default(
        cls,
        plugin_name: str,
        setting_name: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
        post_safe: bool = False,
    ) -> Any:
        """
        Get default setting value from an app plugin.

        :param plugin_name: App plugin name (string, equals "name" in plugin)
        :param setting_name: Setting name (string)
        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether a POST safe value should be returned (bool)
        :return: Setting value
        :raise: ValueError if app plugin is not found
        :raise: KeyError if nothing is found with setting_name
        """
        if plugin_name == APP_NAME:
            s_defs = cls.get_projectroles_defs()
        else:
            s_defs = cls._get_defs(plugin_name=plugin_name)
        if setting_name not in s_defs:
            raise KeyError(
                f'Setting "{setting_name}" not found in app plugin '
                f'"{plugin_name}"'
            )
        s_def = s_defs[setting_name]
        if callable(s_def.default):
            try:
                return s_def.default(project, user)
            except Exception:
                logger.error(
                    f'Error in callable setting "{setting_name}" for plugin '
                    f'"{plugin_name}"'
                )
                return APP_SETTING_DEFAULT_VALUES[s_def.type]
        elif s_def.type == APP_SETTING_TYPE_JSON:
            if s_def.default is None:
                return {}
            if post_safe:
                return json.dumps(s_def.default)
        return s_def.default

    @classmethod
    def get(
        cls,
        plugin_name: str,
        setting_name: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
        post_safe: bool = False,
    ) -> Any:
        """
        Return app setting value for a project or a user. If not set, return
        default.

        :param plugin_name: App plugin name (string, equals "name" in plugin)
        :param setting_name: Setting name (string)
        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether a POST safe value should be returned (bool)
        :return: Value
        :raise: KeyError if nothing is found with setting_name
        """
        if not user or user.is_authenticated:
            try:
                val = AppSetting.objects.get_setting_value(
                    plugin_name, setting_name, project=project, user=user
                )
            except AppSetting.DoesNotExist:
                val = cls.get_default(
                    plugin_name,
                    setting_name,
                    project=project,
                    user=user,
                    post_safe=post_safe,
                )
        else:  # Anonymous user
            val = cls.get_default(
                plugin_name,
                setting_name,
                project=project,
                user=user,
                post_safe=post_safe,
            )
        # Handle post_safe for dict values (JSON)
        if post_safe and isinstance(val, (dict, list)):
            return json.dumps(val)
        return val

    @classmethod
    def get_all_by_scope(
        cls,
        scope: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
        post_safe: bool = False,
    ) -> dict:
        """
        Return all setting values by scope. If a value is not set, return the
        default.

        :param scope: String
        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether POST safe values should be returned (bool)
        :return: Dict
        :raise: ValueError if scope or project and user args are invalid
        """
        PluginAppSettingDef.validate_scope(scope)
        cls._validate_project_and_user(scope, project, user)
        ret = {}
        all_defs = cls.get_all_defs()
        for plugin_name, s_defs in all_defs.items():
            for s_def in [d for d in s_defs.values() if d.scope == scope]:
                ret[f'settings.{plugin_name}.{s_def.name}'] = cls.get(
                    plugin_name, s_def.name, project, user, post_safe
                )
        return ret

    @classmethod
    def get_defaults(
        cls,
        scope: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
        post_safe: bool = False,
    ) -> dict:
        """
        Get all default settings for a scope.

        :param scope: Setting scope (PROJECT, USER or PROJECT_USER)
        :param project: Project object (optional)
        :param user: User object (optional)
        :param post_safe: Whether POST safe values should be returned (bool)
        :return: Dict
        """
        PluginAppSettingDef.validate_scope(scope)
        ret = {}
        app_plugins = plugin_api.get_active_plugins()

        for plugin in app_plugins:
            p_defs = cls.get_definitions(scope, plugin=plugin)
            for s_key in p_defs:
                ret[f'settings.{plugin.name}.{s_key}'] = cls.get_default(
                    plugin.name,
                    s_key,
                    project=project,
                    user=user,
                    post_safe=post_safe,
                )

        p_defs = cls.get_definitions(scope, plugin_name=APP_NAME)
        for s_key in p_defs:
            ret[f'settings.{APP_NAME}.{s_key}'] = cls.get_default(
                APP_NAME,
                s_key,
                project=project,
                user=user,
                post_safe=post_safe,
            )
        return ret

    @classmethod
    def set(
        cls,
        plugin_name: str,
        setting_name: str,
        value: Any,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
        validate: bool = True,
    ) -> bool:
        """
        Set value of an existing project or user settings. Creates the object if
        not found.

        :param plugin_name: App plugin name (string, equals "name" in plugin)
        :param setting_name: Setting name (string)
        :param value: Value to be set
        :param project: Project object (optional)
        :param user: User object (optional)
        :param validate: Validate value (bool, default=True)
        :return: True if changed, False if not changed
        :raise: ValueError if validating and value is not accepted for setting
                type
        :raise: ValueError if neither project nor user are set
        :raise: KeyError if setting name is not found in plugin specification
        """
        s_def = cls.get_definition(name=setting_name, plugin_name=plugin_name)
        cls._validate_project_and_user(s_def.scope, project, user)
        # Check project type
        if project and project.type not in s_def.project_types:
            raise ValueError(
                f'Project type {project.type} not allowed for setting '
                f'{setting_name}'
            )
        # Prevent updating global setting on target site
        if s_def.global_edit:
            if project and project.is_remote():
                raise ValueError(GLOBAL_PROJECT_ERR_MSG)
            if (
                user
                and not project
                and settings.PROJECTROLES_SITE_MODE == SITE_MODE_TARGET
            ):
                raise ValueError(GLOBAL_USER_ERR_MSG)

        # TODO: Refactor and remove repetition
        try:  # Update existing setting
            q_kwargs = {'name': setting_name, 'project': project, 'user': user}
            if not plugin_name == APP_NAME:
                q_kwargs['app_plugin__name'] = plugin_name
            else:
                q_kwargs['app_plugin'] = None
            setting = AppSetting.objects.get(**q_kwargs)
            if cls.compare_value(setting, value):
                return False
            v = (
                cls._get_json_value(value)
                if setting.type == APP_SETTING_TYPE_JSON
                else value
            )
            if validate:
                cls.validate(
                    setting.type,
                    v,
                    s_def.options,
                    project=project,
                    user=user,
                )
            if setting.type == APP_SETTING_TYPE_JSON:
                setting.value_json = v
            else:
                setting.value = v
            setting.save()
            cls._log_set_debug(
                'Set', plugin_name, setting_name, value, project, user
            )
            return True

        except AppSetting.DoesNotExist:  # Create new
            s_type = s_def.type
            if plugin_name == APP_NAME:
                app_plugin_model = None
            else:
                app_plugin = cls._get_app_plugin(plugin_name)
                app_plugin_model = app_plugin.get_model()
            if validate:
                v = (
                    cls._get_json_value(value)
                    if s_type == APP_SETTING_TYPE_JSON
                    else value
                )
                cls.validate(
                    s_type,
                    v,
                    s_def.options,
                    project=project,
                    user=user,
                )
            s_mod = bool(s_def.user_modifiable)
            s_vals = {
                'app_plugin': app_plugin_model,
                'project': project,
                'user': user,
                'name': setting_name,
                'type': s_type,
                'user_modifiable': s_mod,
            }
            if s_type == APP_SETTING_TYPE_JSON:
                s_vals['value_json'] = cls._get_json_value(value)
            else:
                s_vals['value'] = value
            AppSetting.objects.create(**s_vals)
            cls._log_set_debug(
                'Create', plugin_name, setting_name, value, project, user
            )
            return True

    @classmethod
    def is_set(
        cls,
        plugin_name: str,
        setting_name: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ) -> bool:
        """
        Return True if the setting has been set, instead of retrieving the
        default value from the definition.

        NOTE: Also returns True if the current set value equals the default.

        :param plugin_name: App plugin name (string, equals "name" in plugin)
        :param setting_name: Setting name (string)
        :param project: Project object (optional)
        :param user: User object (optional)
        :return: Boolean
        """
        s_def = cls.get_definition(name=setting_name, plugin_name=plugin_name)
        cls._validate_project_and_user(s_def.scope, project, user)
        q_kwargs = {'name': setting_name, 'project': project, 'user': user}
        if not plugin_name == APP_NAME:
            q_kwargs['app_plugin__name'] = plugin_name
        else:
            q_kwargs['app_plugin'] = None
        return AppSetting.objects.filter(**q_kwargs).exists()

    @classmethod
    def delete(
        cls,
        plugin_name: str,
        setting_name: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Delete one or more app setting objects. In case of a PROJECT_USER
        setting, can be used to delete all settings related to project.

        :param plugin_name: App plugin name (string, equals "name" in plugin)
        :param setting_name: Setting name (string)
        :param project: Project object to delete setting from (optional)
        :param user: User object to delete setting from (optional)
        :raise: ValueError with invalid project/user args
        """
        s_def = cls.get_definition(setting_name, plugin_name=plugin_name)
        if s_def.scope != APP_SETTING_SCOPE_PROJECT_USER:
            cls._validate_project_and_user(s_def.scope, project, user)
        elif not project:
            raise ValueError(
                'Project must be set for {} scope settings'.format(
                    APP_SETTING_SCOPE_PROJECT_USER
                )
            )
        q_kwargs = {'name': setting_name}
        if user:
            q_kwargs['user'] = user
        if project:
            q_kwargs['project'] = project
        logger.debug(
            'Delete app setting: {}.{} ({})'.format(
                plugin_name, setting_name, '; '.join(q_kwargs)
            )
        )
        app_settings = AppSetting.objects.filter(**q_kwargs)
        sc = app_settings.count()
        app_settings.delete()
        logger.debug(
            'Deleted {} app setting{}'.format(sc, 's' if sc != 1 else '')
        )

    @classmethod
    def delete_by_scope(
        cls,
        scope: str,
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ):
        """
        Delete all app settings within a given scope for a project and/or user.

        :param scope: Setting scope (string)
        :param project: Project object to delete setting from
        :param user: User object to delete setting from
        :raise: ValueError if scope, project or user are incorrect
        """
        PluginAppSettingDef.validate_scope(scope)
        cls._validate_project_and_user(scope, project, user)
        for plugin_name, app_settings in cls.get_all_defs().items():
            for s_name, s_def in app_settings.items():
                if s_def.scope == scope:
                    cls.delete(plugin_name, s_name, project=project, user=user)

    @classmethod
    def validate(
        cls,
        setting_type: str,
        setting_value: Any,
        setting_options: Union[callable, None, list],
        project: Optional[Project] = None,
        user: Optional[SODARUser] = None,
    ) -> bool:
        """
        Validate setting value according to its type.

        :param setting_type: Setting type (string)
        :param setting_value: Setting value
        :param setting_options: Setting options (can be None)
        :param project: Project object (optional)
        :param user: User object (optional)
        :raise: ValueError if setting_type or setting_value is invalid
        """
        PluginAppSettingDef.validate_type(setting_type)
        cls._validate_value_in_options(
            setting_value, setting_options, project=project, user=user
        )
        # Test callable value
        if callable(setting_value):
            setting_value(project, user)
        else:  # Else validate normal value
            PluginAppSettingDef.validate_value(setting_type, setting_value)
        return True

    @classmethod
    def get_definition(
        cls,
        name: str,
        plugin: Optional[PluginPoint] = None,
        plugin_name: Optional[str] = None,
    ) -> PluginAppSettingDef:
        """
        Return definition for a single app setting, either based on an app name
        or the plugin object.

        :param name: Setting name (string)
        :param plugin: Plugin object or None
        :param plugin_name: Name of the app plugin (string or None)
        :return: PluginAppSettingDef
        :raise: ValueError if neither plugin_name nor plugin are set, or if
                setting is not found in plugin
        """
        defs = cls._get_defs(plugin, plugin_name)
        if name not in defs:
            raise ValueError(
                f'App setting not found in plugin '
                f'"{plugin_name or plugin.name}" with name "{name}"'
            )
        return defs[name]

    @classmethod
    def get_definitions(
        cls,
        scope: Optional[str] = None,
        plugin: Optional[PluginPoint] = None,
        plugin_name: Optional[str] = None,
        user_modifiable: bool = False,
    ) -> dict:
        """
        Return app setting definitions from a plugin, optionally limited by
        scope.

        :param scope: String or None
        :param plugin: Plugin object or None
        :param plugin_name: App plugin name (string, equals "name" in plugin)
        :param user_modifiable: Only return non-superuser modifiable settings if
                                True (boolean)
        :return: Dict
        :raise: ValueError if scope is invalid or if neither plugin_name nor
                plugin are set
        """
        if scope:
            PluginAppSettingDef.validate_scope(scope)
        defs = cls._get_defs(plugin, plugin_name)
        return {
            k: v
            for k, v in defs.items()
            if (not scope or v.scope == scope)
            and (not user_modifiable or v.user_modifiable)
        }

    @classmethod
    def get_projectroles_defs(cls) -> dict:
        """
        Return projectroles settings definitions. If it exists, get value from
        settings.PROJECTROLES_APP_SETTINGS_TEST for testing modifications.

        :return: Dict
        """
        try:
            app_settings = (
                settings.PROJECTROLES_APP_SETTINGS_TEST
                or PROJECTROLES_APP_SETTINGS
            )
        except AttributeError:
            app_settings = PROJECTROLES_APP_SETTINGS
        return {s.name: s for s in app_settings}

    @classmethod
    def get_all_defs(cls) -> dict:
        """
        Return app setting definitions for projectroles and all active app
        plugins in a dictionary with the app name as key.

        :return: Dict
        """
        ret = {APP_NAME: cls.get_projectroles_defs()}
        plugins = (
            []
            + plugin_api.get_active_plugins('project_app')
            + plugin_api.get_active_plugins('site_app')
        )
        for p in plugins:
            ret[p.name] = cls._get_defs(p)
        return ret

    @classmethod
    def compare_value(cls, obj: AppSetting, input_value: Any) -> bool:
        """
        Compare input value to value in an AppSetting object. Return True if
        values match, False if there is a mismatch.

        :param obj: AppSetting object
        :param input_value: Input value (string, int, bool or dict)
        :return: Bool
        """
        if obj.type == APP_SETTING_TYPE_JSON:
            return (
                not obj.value_json and not input_value
            ) or obj.value_json == cls._get_json_value(input_value)
        elif obj.type == APP_SETTING_TYPE_BOOLEAN:
            if isinstance(input_value, str):
                input_value = bool(int(input_value))
            return bool(int(obj.value)) == input_value
        return obj.value == str(input_value)


def get_example_setting_default(
    project: Optional[Project] = None, user: Optional[SODARUser] = None
) -> str:
    """
    Example method for callable default value retrieval for app settings.

    :param project: Project object
    :param user: User object
    :return: String with project and user info or 'No project'
    """
    response = 'N/A'
    if project and user:
        response = f'{project.title}:{user.username}'
    elif project:
        response = str(project.sodar_uuid)
    elif user:
        response = str(user.sodar_uuid)
    return response


def get_example_setting_options(
    project: Optional[Project] = None, user: Optional[SODARUser] = None
) -> list[tuple]:
    """
    Example method for callable option list retrieval for app settings.

    :param project: Project object
    :param user: User object
    :return: List of tuples for ChoiceField
    """
    ret = [('N/A', 'No project or user for callable'), 'Example string option']
    if project and user:
        ret.append(
            (
                str(project.sodar_uuid),
                f'Project UUID {project.sodar_uuid} by {user.username}',
            )
        )
    elif project:
        ret.append(
            (
                str(project.sodar_uuid),
                f'Project UUID: {project.sodar_uuid}',
            )
        )
    elif user:
        ret.append((str(user.sodar_uuid), f'User UUID: {user.sodar_uuid}'))
    return ret



"""REST API view permission tests for the sodarcache app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.plugins import PluginAPI
from projectroles.tests.test_permissions_api import ProjectAPIPermissionTestBase

from sodarcache.tests.test_views_api import (
    TEST_APP_NAME,
    ITEM_NAME,
    DATA_KEY,
    DATA_VAL,
)
from sodarcache.views_api import (
    SODARCACHE_API_MEDIA_TYPE,
    SODARCACHE_API_DEFAULT_VERSION,
)


plugin_api = PluginAPI()


class SodarcacheAPIPermissionTestBase(ProjectAPIPermissionTestBase):
    """Base class for sodarcache REST API view permission tests"""

    media_type = SODARCACHE_API_MEDIA_TYPE
    api_version = SODARCACHE_API_DEFAULT_VERSION

    def setUp(self):
        super().setUp()
        self.cache_backend = plugin_api.get_backend_api('sodar_cache')


class TestCacheItemRetrieveAPIView(SodarcacheAPIPermissionTestBase):
    """Test CacheItemRetrieveAPIView permissions"""

    def setUp(self):
        super().setUp()
        # Init cache item
        self.item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=TEST_APP_NAME,
            user=self.user_owner,
            name=ITEM_NAME,
            data={DATA_KEY: DATA_VAL},
        )
        self.url = reverse(
            'sodarcache:api_retrieve',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test CacheItemRetrieveAPIView GET"""
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)


class TestCacheItemDateRetrieveAPIView(SodarcacheAPIPermissionTestBase):
    """Test CacheItemDateRetrieveAPIView permissions"""

    def setUp(self):
        super().setUp()
        # Init cache item
        self.item = self.cache_backend.set_cache_item(
            project=self.project,
            app_name=TEST_APP_NAME,
            user=self.user_owner,
            name=ITEM_NAME,
            data={DATA_KEY: DATA_VAL},
        )
        self.url = reverse(
            'sodarcache:api_retrieve_date',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_guest_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
            self.user_guest,
        ]
        self.bad_users = [
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_get(self):
        """Test CacheItemDateRetrieveAPIView GET"""
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.anonymous, 200)

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)
        self.assert_response_api(self.url, self.good_users, 200, knox=True)
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(self.url, self.user_no_roles, 200)

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(self.url, self.good_users, 200)
        self.assert_response_api(self.url, self.bad_users, 403)
        self.assert_response_api(self.url, self.anonymous, 401)


class TestCacheItemSetAPIView(SodarcacheAPIPermissionTestBase):
    """Test CacheItemSetAPIView permissions"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'sodarcache:api_set',
            kwargs={
                'project': self.project.sodar_uuid,
                'app_name': TEST_APP_NAME,
                'item_name': ITEM_NAME,
            },
        )
        self.post_data = {'data': {DATA_KEY: DATA_VAL}}
        self.good_users = [
            self.superuser,
            self.user_owner_cat,
            self.user_delegate_cat,
            self.user_contributor_cat,
            self.user_owner,
            self.user_delegate,
            self.user_contributor,
        ]
        self.bad_users = [
            self.user_guest_cat,
            self.user_viewer_cat,
            self.user_finder_cat,
            self.user_guest,
            self.user_viewer,
            self.user_no_roles,
        ]

    def test_post(self):
        """Test CacheItemSetAPIView POST"""
        self.assert_response_api(
            self.url, self.good_users, 200, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.bad_users, 403, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            self.good_users,
            200,
            method='POST',
            data=self.post_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            data=self.post_data,
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )

    def test_get_archive(self):
        """Test GET with archived project"""
        self.project.set_archive()
        self.assert_response_api(
            self.url, self.good_users, 200, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.bad_users, 403, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url,
            self.good_users,
            200,
            method='POST',
            data=self.post_data,
            knox=True,
        )
        self.project.set_public_access(self.role_guest)
        self.assert_response_api(
            self.url,
            self.user_no_roles,
            403,
            method='POST',
            data=self.post_data,
        )

    def test_get_read_only(self):
        """Test GET with site read-only mode"""
        self.set_site_read_only()
        self.assert_response_api(
            self.url, self.good_users, 200, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.bad_users, 403, method='POST', data=self.post_data
        )
        self.assert_response_api(
            self.url, self.anonymous, 401, method='POST', data=self.post_data
        )



"""
SODAR project and role management
"""

from . import _version  # noqa

__version__ = _version.get_versions()['version']

default_app_config = (
    'projectroles.apps.ProjectrolesConfig'  # pylint: disable=invalid-name
)



"""Permissions for the tokens app"""

import rules

from django.conf import settings

# Projectroles dependency
from projectroles import rules as pr_rules  # To access common predicates
from projectroles.models import RoleAssignment


# Predicates -------------------------------------------------------------------


@rules.predicate
def can_create_token(user):
    if (
        not user.is_superuser
        and getattr(settings, 'TOKENS_CREATE_PROJECT_USER_RESTRICT', False)
        and RoleAssignment.objects.filter(user=user).count() == 0
    ):
        return False
    return True


# Permissions ------------------------------------------------------------------

# View tokens list
rules.add_perm('tokens.view_list', rules.is_authenticated)

# Create token
rules.add_perm(
    'tokens.create',
    rules.is_authenticated & pr_rules.is_site_writable & can_create_token,
)

# Delete token
rules.add_perm(
    'tokens.delete', rules.is_authenticated & pr_rules.is_site_writable
)



"""Access rules for the sodarcache app"""

import rules

# Projectroles dependency
from projectroles import rules as pr_rules  # To access common predicates


# Predicates -------------------------------------------------------------


# If we need to assign new predicates, we do it here


# Rules ------------------------------------------------------------------


# Rules should not be needed, use permissions for user rights


# Permissions ------------------------------------------------------------


# Allow viewing cached data
rules.add_perm(
    'sodarcache.get_cache_value',
    pr_rules.is_project_owner
    | pr_rules.is_project_delegate
    | pr_rules.is_project_contributor
    | pr_rules.is_project_guest,
)

# Allow setting cached data
rules.add_perm(
    'sodarcache.set_cache_value',
    pr_rules.is_project_owner
    | pr_rules.is_project_delegate
    | pr_rules.is_project_contributor,
)



"""Tests for view permissions in the appalerts app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions import SiteAppPermissionTestBase

from appalerts.tests.test_models import AppAlertMixin


class AppalertsPermissionTestBase(AppAlertMixin, SiteAppPermissionTestBase):
    """Base test class for appalerts view permission tests"""

    def setUp(self):
        super().setUp()
        # Create user with no alerts
        self.no_alert_user = self.make_user('no_alert_user')
        # Create alert
        self.alert = self.make_app_alert(
            user=self.regular_user, url=reverse('home')
        )


class TestAppAlertListView(AppalertsPermissionTestBase):
    """Permission tests for AppAlertListView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('appalerts:list')

    def test_get(self):
        """Test AppAlertListView GET"""
        good_users = [
            self.superuser,
            self.regular_user,
            self.no_alert_user,
        ]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, self.anonymous, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(self.url, self.anonymous, 302)


class TestAppAlertRedirectView(AppalertsPermissionTestBase):
    """Permission tests for AppAlertLinkRedirectView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'appalerts:redirect', kwargs={'appalert': self.alert.sodar_uuid}
        )
        self.bad_redirect_url = reverse('appalerts:list')

    def test_get(self):
        """Test AppAlertLinkRedirectView GET"""
        good_users = [self.regular_user]
        bad_users = [self.superuser, self.no_alert_user, self.anonymous]
        self.assert_response(
            self.url, good_users, 302, redirect_user=reverse('home')
        )
        self.assert_response(
            self.url, bad_users, 302, redirect_user=self.bad_redirect_url
        )

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(
            self.url, self.anonymous, 302, redirect_user=self.bad_redirect_url
        )


class TestAppAlertStatusAjaxView(AppalertsPermissionTestBase):
    """Permission tests for AppAlertStatusAjaxView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('appalerts:ajax_status')

    def test_get(self):
        """Test AppAlertStatusAjaxView GET"""
        good_users = [self.superuser, self.regular_user, self.no_alert_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, self.anonymous, 403)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        self.assert_response(self.url, self.anonymous, 403)


class TestAppAlertDismissAjaxView(AppalertsPermissionTestBase):
    """Permission tests for AppAlertDismissAjaxView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'appalerts:ajax_dismiss', kwargs={'appalert': self.alert.sodar_uuid}
        )

    def test_post(self):
        """Test AppAlertDismissAjaxView POST"""
        good_users = [self.regular_user]
        bad_users = [self.superuser, self.no_alert_user]
        self.assert_response(self.url, good_users, 200, method='POST')
        self.assert_response(self.url, bad_users, 404, method='POST')
        self.assert_response(self.url, self.anonymous, 403, method='POST')

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_post_anon(self):
        """Test POST with anonymous access"""
        self.assert_response(self.url, self.anonymous, 403, method='POST')



"""
URL configuration for the bgjobs app.

Note that the job detail view is implemented
"""

from django.urls import path

from bgjobs import views

app_name = 'bgjobs'

urls_ui_project = [
    # List jobs that the user has access to
    path(
        route='list/<uuid:project>',
        view=views.ProjectBackgroundJobView.as_view(),
        name='list',
    ),
    # Clear jobs in project owned by the current user
    path(
        route='clear/own/<uuid:project>',
        view=views.BackgroundJobClearOwnView.as_view(),
        name='clear_own',
    ),
    # Clear jobs in project regardless of the user
    path(
        route='clear/all/<uuid:project>',
        view=views.BackgroundJobClearAllView.as_view(),
        name='clear_all',
    ),
]

urls_ui_site = [
    # List site background jobs
    path(
        route='list',
        view=views.SiteBackgroundJobView.as_view(),
        name='site_list',
    ),
]

urlpatterns = urls_ui_project + urls_ui_site
